import pytest

from helpers import gdal_dataset_of_region
from yirgacheffe.window import Area, Window
from yirgacheffe.layers import ConstantLayer, RasterLayer


def test_find_union_empty_list() -> None:
    with pytest.raises(ValueError):
        _ = RasterLayer.find_union([])

def test_find_union_single_item() -> None:
    layer = RasterLayer(gdal_dataset_of_region(Area(-10, 10, 10, -10), 0.02))
    union = RasterLayer.find_union([layer])
    assert union == layer.area

def test_find_union_same() -> None:
    layers = [
        RasterLayer(gdal_dataset_of_region(Area(-10, 10, 10, -10), 0.02)),
        RasterLayer(gdal_dataset_of_region(Area(-10, 10, 10, -10), 0.02))
    ]
    union = RasterLayer.find_union(layers)
    assert union == layers[0].area

def test_find_union_subset() -> None:
    layers = [
        RasterLayer(gdal_dataset_of_region(Area(-10, 10, 10, -10), 0.02)),
        RasterLayer(gdal_dataset_of_region(Area(-1, 1, 1, -1), 0.02))
    ]
    union = RasterLayer.find_union(layers)
    assert union == layers[0].area

def test_find_union_overlap() -> None:
    layers = [
        RasterLayer(gdal_dataset_of_region(Area(-10, 10, 10, -10), 0.02)),
        RasterLayer(gdal_dataset_of_region(Area(-15, 15, -5, -5), 0.02))
    ]
    union = RasterLayer.find_union(layers)
    assert union == Area(-15, 15, 10, -10)

def test_find_union_distinct() -> None:
    layers = [
        RasterLayer(gdal_dataset_of_region(Area(-110, 10, -100, -10), 0.02)),
        RasterLayer(gdal_dataset_of_region(Area(100, 10, 110, -10), 0.02))
    ]
    union = RasterLayer.find_union(layers)
    assert union == Area(-110, 10, 110, -10)

def test_find_union_with_null() -> None:
    layers = [
        RasterLayer(gdal_dataset_of_region(Area(-10, 10, 10, -10), 0.02)),
        ConstantLayer(0.0)
    ]
    union = RasterLayer.find_union(layers)
    assert union == layers[1].area

def test_find_union_different_pixel_pitch() -> None:
    layers = [
        RasterLayer(gdal_dataset_of_region(Area(-10, 10, 10, -10), 0.02)),
        RasterLayer(gdal_dataset_of_region(Area(-15, 15, -5, -5), 0.01))
    ]
    with pytest.raises(ValueError):
        _ = RasterLayer.find_union(layers)

@pytest.mark.parametrize("scale", [0.01, 0.02, 0.03, 0.04, 0.05, 0.06, 0.07, 0.08, 0.09])
def test_set_union_self(scale) -> None:
    layer = RasterLayer(gdal_dataset_of_region(Area(-10, 10, 10, -10), scale))
    old_window = layer.window

    # note that the area we passed to gdal_dataset_of_region isn't pixel aligned, so we must
    # use the area from loading the dataset
    layer.set_window_for_union(layer.area)
    assert layer.window == old_window

    # reset should not do much here
    layer.reset_window()
    assert layer.window == old_window

@pytest.mark.parametrize(
    "left_padding,right_padding,top_padding,bottom_padding",
    [
        (1, 1, 0, 0),
        (0, 1, 0, 0),
        (1, 0, 0, 0),
        (0, 0, 0, 0),
        (1, 1, 1, 0),
        (0, 1, 1, 0),
        (1, 0, 1, 0),
        (0, 0, 1, 0),
        (1, 1, 0, 1),
        (0, 1, 0, 1),
        (1, 0, 0, 1),
        (0, 0, 0, 1),
        (1, 1, 1, 1),
        (0, 1, 1, 1),
        (1, 0, 1, 1),
        (0, 0, 1, 1),
    ]
)
def test_set_union_superset(left_padding: int, right_padding: int, top_padding: int, bottom_padding: int) -> None:

    pixel_density = 0.02
    origin_area = Area(-1, 1, 1, -1)

    layer = RasterLayer(gdal_dataset_of_region(origin_area, pixel_density))
    assert layer.window == Window(0, 0, 100, 100)

    # The make_dataset... function fills rows with the yoffset, and so the first row
    # will be 0s, matching our padding value, so we use the second row here
    origin_before_pixel = layer.read_array(0, 1, 100, 1)
    assert list(origin_before_pixel[0]) == ([1,] * 100)

    # Superset only extends on both sides
    superset = Area(-1 - left_padding, 1 + top_padding, 1 + right_padding, -1 - bottom_padding)
    layer.set_window_for_union(superset)
    assert layer.window == Window(
        (0 - left_padding) / pixel_density,
        (0 - top_padding) / pixel_density,
        (2 + left_padding + right_padding) / pixel_density,
        (2 + top_padding + bottom_padding) / pixel_density,
    )

    origin_after_pixel = layer.read_array(
        0,
        1 + int(top_padding / pixel_density),
        100 + int((left_padding + right_padding) / pixel_density),
        1
    )
    assert list(origin_after_pixel[0]) == (
        [0,] * int(left_padding / pixel_density)) +\
        list(origin_before_pixel[0]) +\
        ([0,] * int(right_padding / pixel_density)
    )

    layer.reset_window()
    assert layer.window == Window(0, 0, 100, 100)



import math
import sys
from collections import namedtuple
from dataclasses import dataclass
from typing import List, Optional

PixelScale = namedtuple('PixelScale', ['xstep', 'ystep'])

@dataclass
class Area:
    left: float
    top: float
    right: float
    bottom: float

    def __hash__(self):
        return (self.left, self.top, self.right, self.bottom).__hash__()

    def __eq__(self, other) -> bool:
        return math.isclose(self.left, other.left, abs_tol=1e-09) and \
            math.isclose(self.right, other.right, abs_tol=1e-09) and \
            math.isclose(self.top, other.top, abs_tol=1e-09) and \
            math.isclose(self.bottom, other.bottom, abs_tol=1e-09)

    def grow(self, offset: float):
        return Area(
            left=self.left - offset,
            top=self.top + offset,
            right=self.right + offset,
            bottom=self.bottom - offset
        )

    def overlaps(self, other) -> bool:
        return (
            (self.left <= other.left <= self.right) or
            (self.left <= other.right <= self.right) or
            (other.left <= self.left <= other.right) or
            (other.left <= self.right <= other.right)
        ) and (
            (self.top >= other.top >= self.bottom) or
            (self.top >= other.bottom >= self.bottom) or
            (other.top >= self.top >= other.bottom) or
            (other.top >= self.bottom >= other.bottom)
        )

@dataclass
class Window:
    xoff: int
    yoff: int
    xsize: int
    ysize: int

    @property
    def as_array_args(self):
        return (self.xoff, self.yoff, self.xsize, self.ysize)

    def __lt__(self, other) -> bool:
        return (self.xsize < other.xsize) and \
            (self.ysize < other.ysize) and \
            (self.xoff >= other.xoff) and \
            (self.yoff >= other.yoff) and \
            ((self.xoff + self.xsize) <= (other.xoff + other.xsize)) and \
            ((self.yoff + self.ysize) <= (other.yoff + other.ysize))

    def __gt__(self, other) -> bool:
        return (self.xsize > other.xsize) and \
            (self.ysize > other.ysize) and \
            (self.xoff <= other.xoff) and \
            (self.yoff <= other.yoff) and \
            ((self.xoff + self.xsize) >= (other.xoff + other.xsize)) and \
            ((self.yoff + self.ysize) >= (other.yoff + other.ysize))

    def __le__(self, other) -> bool:
        return (self.xsize <= other.xsize) and \
            (self.ysize <= other.ysize) and \
            (self.xoff >= other.xoff) and \
            (self.yoff >= other.yoff) and \
            ((self.xoff + self.xsize) <= (other.xoff + other.xsize)) and \
            ((self.yoff + self.ysize) <= (other.yoff + other.ysize))

    def __ge__(self, other) -> bool:
        return (self.xsize >= other.xsize) and \
            (self.ysize >= other.ysize) and \
            (self.xoff <= other.xoff) and \
            (self.yoff <= other.yoff) and \
            ((self.xoff + self.xsize) >= (other.xoff + other.xsize)) and \
            ((self.yoff + self.ysize) >= (other.yoff + other.ysize))

    def grow(self, pixels: int):
        return Window(
            xoff=self.xoff - pixels,
            yoff=self.xoff - pixels,
            xsize=self.xsize + (2 * pixels),
            ysize=self.ysize + (2 * pixels),
        )

    @staticmethod
    def find_intersection(windows: List) -> "Window":
        if not windows:
            raise ValueError("Expected list of windows")
        # This isn't very pythonic, as it was originally written, but
        # this method gets called a lot (by every Layer.read_array), so not looping
        # over the window list multiple times halves the performance cost (as
        # measured with cProfile).
        left, top, right, bottom = -sys.maxsize, -sys.maxsize, sys.maxsize, sys.maxsize
        for window in windows:
            left = max(left, window.xoff)
            top = max(top, window.yoff)
            right = min(right, window.xoff + window.xsize)
            bottom = min(bottom, window.yoff + window.ysize)
        if (left >= right) or (top >= bottom):
            raise ValueError('No intersection possible')
        return Window(
            left,
            top,
            right - left,
            bottom - top
        )

    @staticmethod
    def find_intersection_no_throw(windows: List) -> Optional["Window"]:
        if not windows:
            raise ValueError("Expected list of windows")
        # This isn't very pythonic, as it was originally written, but
        # this method gets called a lot (by every Layer.read_array), so not looping
        # over the window list multiple times halves the performance cost (as
        # measured with cProfile).
        left, top, right, bottom = -sys.maxsize, -sys.maxsize, sys.maxsize, sys.maxsize
        for window in windows:
            left = max(left, window.xoff)
            top = max(top, window.yoff)
            right = min(right, window.xoff + window.xsize)
            bottom = min(bottom, window.yoff + window.ysize)
        if (left >= right) or (top >= bottom):
            return None
        return Window(
            left,
            top,
            right - left,
            bottom - top
        )



import sys
from typing import List

import pytest

from yirgacheffe.window import PixelScale
from yirgacheffe.rounding import almost_equal, are_pixel_scales_equal_enough, round_up_pixels,\
    round_down_pixels, MINIMAL_DEGREE_OF_INTEREST

@pytest.mark.parametrize("lval,rval,expected",
    [
        (1.0, 1.0, True),
        (1.0, 1.1, False),
        (sys.float_info.epsilon / 2, 0.0, True),
        (sys.float_info.epsilon * 2, 0.0, False),
    ]
)
def test_almost_equal(lval, rval, expected):
    assert almost_equal(lval, rval) == expected

# The pixel scale here comes from the jung dataset, which is 400752 pixles
# wide, or 100M per pixel at the equator roughly.
@pytest.mark.parametrize("pixels,scale,expected",
    [
        (8033.000000000001, 0.0008983152841195215, 8033), # actual seen value
        (8033.001, 0.0008983152841195215, 8033),          # obvious just below test case
        (8033.01, 0.0008983152841195215, 8034),           # obvious just above test case
        (8032.999999999999, 0.0008983152841195215, 8033), # actual seen value
    ]
)
def test_pixel_rounding_up(pixels: float, scale: float, expected: int) -> None:
    assert round_up_pixels(pixels, scale) == expected

@pytest.mark.parametrize("pixels,scale,expected",
    [
        (55.99999999999926, 0.0008983152841195215, 56), # actual seen value
        (55.998, 0.0008983152841195215, 56),            # obvious just below test case
        (55.98, 0.0008983152841195215, 55),             # obvious just above test case
        (55.000000000001, 0.0008983152841195215, 55),   # actual seen value
    ]
)
def test_pixel_rounding_down(pixels: float, scale: float, expected: int) -> None:
    assert round_down_pixels(pixels, scale) == expected

@pytest.mark.parametrize("pixel_scales,expected",
    [
        (
            [],
            True
        ),
        (
            [
                PixelScale(0.1, 0.1),
            ],
            True
        ),
        (
            [
                PixelScale(0.1, 0.1),
                None,
            ],
            True
        ),
        (
            [
                PixelScale(0.1, 0.1),
                PixelScale(0.1, 0.1),
            ],
            True
        ),
        (
            [
                PixelScale(0.1, 0.1),
                PixelScale(0.1 + (MINIMAL_DEGREE_OF_INTEREST / 2), 0.1 + (MINIMAL_DEGREE_OF_INTEREST / 2)),
            ],
            True
        ),
        (
            [
                PixelScale(0.1, 0.1),
                PixelScale(0.1 + (MINIMAL_DEGREE_OF_INTEREST * 2), 0.1 + (MINIMAL_DEGREE_OF_INTEREST * 2)),
            ],
            False
        ),
    ]
)
def test_pixel_scale_comparison(pixel_scales: List[PixelScale], expected: bool) -> None:
    assert are_pixel_scales_equal_enough(pixel_scales) == expected



from osgeo import gdal
try:
    from importlib import metadata
    __version__ = metadata.version(__name__)
except ModuleNotFoundError:
    __version__ = "unknown"

gdal.UseExceptions()

# I don't really want this here, but it's just too useful having it exposed
WGS_84_PROJECTION = 'GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.257223563,'\
    'AUTHORITY["EPSG","7030"]],AUTHORITY["EPSG","6326"]],PRIMEM["Greenwich",0],'\
    'UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AXIS["Latitude",NORTH],'\
    'AXIS["Longitude",EAST],AUTHORITY["EPSG","4326"]]'

# For legacy reasons [facepalm]
WSG_84_PROJECTION = WGS_84_PROJECTION




from typing import Any, List, Optional, Tuple

from ..operators import LayerMathMixin
from ..rounding import almost_equal, are_pixel_scales_equal_enough, round_up_pixels, round_down_pixels
from ..window import Area, PixelScale, Window

class YirgacheffeLayer(LayerMathMixin):
    """The common base class for the different layer types. Most still inherit from RasterLayer as deep down
    they end up as pixels, but this is a start to make other layers that don't need to rasterize not have
    to carry all that baggage."""

    def __init__(self,
        area: Area,
        pixel_scale: Optional[PixelScale],
        projection: str,
        name: Optional[str] = None
    ):
        self._pixel_scale = pixel_scale
        self._underlying_area = area
        self._active_area = area
        self._projection = projection
        self._window: Optional[Window] = None
        self.name = name

        self.reset_window()

    def close(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    def datatype(self) -> int:
        raise NotImplementedError("Must be overridden by subclass")

    @property
    def projection(self) -> str:
        return self._projection

    @property
    def pixel_scale(self) -> Optional[PixelScale]:
        return self._pixel_scale

    @property
    def area(self) -> Area:
        return self._active_area

    @property
    def window(self) -> Window:
        if self._window is None:
            raise AttributeError("Layer has no window")
        return self._window

    @staticmethod
    def find_intersection(layers: List) -> Area:
        if not layers:
            raise ValueError("Expected list of layers")

        # This only makes sense (currently) if all layers
        # have the same pixel pitch (modulo desired accuracy)
        if not are_pixel_scales_equal_enough([x.pixel_scale for x in layers]):
            raise ValueError("Not all layers are at the same pixel scale")

        intersection = Area(
            left=max(x._underlying_area.left for x in layers),
            top=min(x._underlying_area.top for x in layers),
            right=min(x._underlying_area.right for x in layers),
            bottom=max(x._underlying_area.bottom for x in layers)
        )
        if (intersection.left >= intersection.right) or (intersection.bottom >= intersection.top):
            raise ValueError('No intersection possible')
        return intersection

    @staticmethod
    def find_union(layers: List) -> Area:
        if not layers:
            raise ValueError("Expected list of layers")

        # This only makes sense (currently) if all layers
        # have the same pixel pitch (modulo desired accuracy)
        if not are_pixel_scales_equal_enough([x.pixel_scale for x in layers]):
            raise ValueError("Not all layers are at the same pixel scale")

        return Area(
            left=min(x._underlying_area.left for x in layers),
            top=max(x._underlying_area.top for x in layers),
            right=max(x._underlying_area.right for x in layers),
            bottom=min(x._underlying_area.bottom for x in layers)
        )

    @property
    def geo_transform(self) -> Tuple[float, float, float, float, float, float]:
        if self._pixel_scale is None:
            raise ValueError("No geo transform for layers without explicit pixel scale")
        return (
            self._active_area.left, self._pixel_scale.xstep, 0.0,
            self._active_area.top, 0.0, self._pixel_scale.ystep
        )

    def check_pixel_scale(self, scale: PixelScale) -> bool:
        our_scale = self.pixel_scale
        if our_scale is None:
            raise ValueError("No check for layers without explicit pixel scale")
        return almost_equal(our_scale.xstep, scale.xstep) and \
            almost_equal(our_scale.ystep, scale.ystep)

    def set_window_for_intersection(self, new_area: Area) -> None:
        if self._pixel_scale is None:
            raise ValueError("Can not set Window without explicit pixel scale")

        new_window = Window(
            xoff=round_down_pixels((new_area.left - self._underlying_area.left) / self._pixel_scale.xstep,
                self._pixel_scale.xstep),
            yoff=round_down_pixels((self._underlying_area.top - new_area.top) / (self._pixel_scale.ystep * -1.0),
                self._pixel_scale.ystep * -1.0),
            xsize=round_up_pixels(
                (new_area.right - new_area.left) / self._pixel_scale.xstep,
                self._pixel_scale.xstep
            ),
            ysize=round_up_pixels(
                (new_area.top - new_area.bottom) / (self._pixel_scale.ystep * -1.0),
                (self._pixel_scale.ystep * -1.0)
            ),
        )
        if (new_window.xoff < 0) or (new_window.yoff < 0):
            raise ValueError('Window has negative offset')
        # If there is an underlying raster for this layer, do a sanity check
        try:
            if ((new_window.xoff + new_window.xsize) > self._raster_xsize) or \
                ((new_window.yoff + new_window.ysize) > self._raster_ysize):
                raise ValueError(f'Window is bigger than dataset: raster is {self._raster_xsize}x{self._raster_ysize}'\
                    f', new window is {new_window.xsize - new_window.xoff}x{new_window.ysize - new_window.yoff}')
        except AttributeError:
            pass
        self._window = new_window
        self._active_area = new_area

    def set_window_for_union(self, new_area: Area) -> None:
        if self._pixel_scale is None:
            raise ValueError("Can not set Window without explicit pixel scale")

        new_window = Window(
            xoff=round_down_pixels((new_area.left - self._underlying_area.left) / self._pixel_scale.xstep,
                self._pixel_scale.xstep),
            yoff=round_down_pixels((self._underlying_area.top - new_area.top) / (self._pixel_scale.ystep * -1.0),
                self._pixel_scale.ystep * -1.0),
            xsize=round_up_pixels(
                (new_area.right - new_area.left) / self._pixel_scale.xstep,
                self._pixel_scale.xstep
            ),
            ysize=round_up_pixels(
                (new_area.top - new_area.bottom) / (self._pixel_scale.ystep * -1.0),
                (self._pixel_scale.ystep * -1.0)
            ),
        )
        if (new_window.xoff > 0) or (new_window.yoff > 0):
            raise ValueError('Window has positive offset')
        # If there is an underlying raster for this layer, do a sanity check
        try:
            if ((new_window.xsize - new_window.xoff) < self._raster_xsize) or \
                ((new_window.ysize - new_window.yoff) < self._raster_ysize):
                raise ValueError(f'Window is smaller than dataset: raster is {self._raster_xsize}x{self._raster_ysize}'\
                    f', new window is {new_window.xsize - new_window.xoff}x{new_window.ysize - new_window.yoff}')
        except AttributeError:
            pass
        self._window = new_window
        self._active_area = new_area

    def reset_window(self):
        self._active_area = self._underlying_area
        if self._pixel_scale:
            abs_xstep, abs_ystep = abs(self._pixel_scale.xstep), abs(self._pixel_scale.ystep)
            self._window = Window(
                xoff=0,
                yoff=0,
                xsize=round_up_pixels((self.area.right - self.area.left) / self._pixel_scale.xstep, abs_xstep),
                ysize=round_up_pixels((self.area.bottom - self.area.top) / self._pixel_scale.ystep, abs_ystep),
            )

    def offset_window_by_pixels(self, offset: int):
        """Grows (if pixels is positive) or shrinks (if pixels is negative) the window for the layer."""
        if offset == 0:
            return

        if offset < 0:
            if (offset * -2 >= self.window.xsize) or (offset * -2 >= self.window.ysize):
                raise ValueError(f"Can not shrink window by {offset}, would make size 0 or fewer pixels.")

        new_window = Window(
            xoff=self.window.xoff - offset,
            yoff=self.window.yoff - offset,
            xsize=self.window.xsize + (2 * offset),
            ysize=self.window.ysize + (2 * offset),
        )
        scale = self.pixel_scale
        if scale is None:
            raise ValueError("Can not offset Window without explicit pixel scale")

        # Note we can't assume that we weren't already on an intersection when making the offset!
        # But remember that window is always relative to underlying area, and new_window
        # here is based off the existing window
        new_left = self._underlying_area.left + (new_window.xoff * scale.xstep)
        new_top = self._underlying_area.top + (new_window.yoff * scale.ystep)
        new_area = Area(
            left=new_left,
            top=new_top,
            right=new_left + (new_window.xsize * scale.xstep),
            bottom=new_top + (new_window.ysize * scale.ystep)
        )
        self._window = new_window
        self._active_area = new_area

    def read_array_with_window(self, _x: int, _y: int, _xsize: int, _ysize: int, window: Window) -> Any:
        raise NotImplementedError("Must be overridden by subclass")

    def read_array_for_area(self, target_area: Area, x: int, y: int, width: int, height: int) -> Any:

        target_window = Window(
            xoff=round_down_pixels((target_area.left - self._underlying_area.left) / self._pixel_scale.xstep,
                self._pixel_scale.xstep),
            yoff=round_down_pixels((self._underlying_area.top - target_area.top) / (self._pixel_scale.ystep * -1.0),
                self._pixel_scale.ystep * -1.0),
            xsize=round_up_pixels(
                (target_area.right - target_area.left) / self._pixel_scale.xstep,
                self._pixel_scale.xstep
            ),
            ysize=round_up_pixels(
                (target_area.top - target_area.bottom) / (self._pixel_scale.ystep * -1.0),
                (self._pixel_scale.ystep * -1.0)
            ),
        )
        return self.read_array_with_window(x, y, width, height, target_window)

    def read_array(self, x: int, y: int, width: int, height: int) -> Any:
        return self.read_array_with_window(x, y, width, height, self.window)

    def latlng_for_pixel(self, x_coord: int, y_coord: int) -> Tuple[float,float]:
        """Get geo coords for pixel. This is relative to the set view window."""
        if "WGS 84" not in self.projection:
            raise NotImplementedError("Not yet supported for other projections")
        pixel_scale = self.pixel_scale
        if pixel_scale is None:
            raise ValueError("Layer has no pixel scale")
        return (
            (y_coord * pixel_scale.ystep) + self.area.top,
            (x_coord * pixel_scale.xstep) + self.area.left
        )

    def pixel_for_latlng(self, lat: float, lng: float) -> Tuple[int,int]:
        """Get pixel for geo coords. This is relative to the set view window.
        Result is rounded down to nearest pixel."""
        if "WGS 84" not in self.projection:
            raise NotImplementedError("Not yet supported for other projections")
        pixel_scale = self.pixel_scale
        if pixel_scale is None:
            raise ValueError("Layer has no pixel scale")
        return (
            round_down_pixels((lng - self.area.left) / pixel_scale.xstep, abs(pixel_scale.xstep)),
            round_down_pixels((lat - self.area.top) / pixel_scale.ystep, abs(pixel_scale.ystep)),
        )



import logging
import multiprocessing
import sys
import time
import types
from enum import Enum
from multiprocessing import Semaphore, Process
from multiprocessing.managers import SharedMemoryManager
from typing import Optional

import numpy as np
from osgeo import gdal
from dill import dumps, loads

from . import constants
from .rounding import are_pixel_scales_equal_enough, round_up_pixels, round_down_pixels
from .window import Area, PixelScale, Window
from .backends import backend
from .backends.enumeration import operators as op

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)

class WindowOperation(Enum):
    NONE = 1
    UNION = 2
    INTERSECTION = 3
    LEFT = 4
    RIGHT = 5

class LayerConstant:
    def __init__(self, val):
        self.val = val

    def __str__(self):
        return str(self.val)

    def _eval(self, _area, _index, _step, _target_window):
        return self.val


class LayerMathMixin:

    def __add__(self, other):
        return LayerOperation(self, op.ADD, other, window_op=WindowOperation.UNION)

    def __sub__(self, other):
        return LayerOperation(self, op.SUB, other, window_op=WindowOperation.UNION)

    def __mul__(self, other):
        return LayerOperation(self, op.MUL, other, window_op=WindowOperation.INTERSECTION)

    def __truediv__(self, other):
        return LayerOperation(self, op.TRUEDIV, other, window_op=WindowOperation.INTERSECTION)

    def __floordiv__(self, other):
        return LayerOperation(self, op.FLOORDIV, other, window_op=WindowOperation.INTERSECTION)

    def __mod__(self, other):
        return LayerOperation(self, op.REMAINDER, other, window_op=WindowOperation.INTERSECTION)

    def __pow__(self, other):
        return LayerOperation(self, op.POW, other, window_op=WindowOperation.UNION)

    def __eq__(self, other):
        return LayerOperation(self, op.EQ, other, window_op=WindowOperation.INTERSECTION)

    def __ne__(self, other):
        return LayerOperation(self, op.NE, other, window_op=WindowOperation.UNION)

    def __lt__(self, other):
        return LayerOperation(self, op.LT, other, window_op=WindowOperation.UNION)

    def __le__(self, other):
        return LayerOperation(self, op.LE, other, window_op=WindowOperation.UNION)

    def __gt__(self, other):
        return LayerOperation(self, op.GT, other, window_op=WindowOperation.UNION)

    def __ge__(self, other):
        return LayerOperation(self, op.GE, other, window_op=WindowOperation.UNION)

    def __and__(self, other):
        return LayerOperation(self, op.AND, other, window_op=WindowOperation.INTERSECTION)

    def __or__(self, other):
        return LayerOperation(self, op.OR, other, window_op=WindowOperation.UNION)

    def _eval(self, area, index, step, target_window=None):
        try:
            window = self.window if target_window is None else target_window
            return self.read_array_for_area(area, 0, index, window.xsize, step)
        except AttributeError:
            return self.read_array_for_area(area, 0, index, target_window.xsize if target_window else 1, step)

    def nan_to_num(self, nan=0, posinf=None, neginf=None):
        return LayerOperation(
            self,
            op.NAN_TO_NUM,
            window_op=WindowOperation.NONE,
            copy=False,
            nan=nan,
            posinf=posinf,
            neginf=neginf,
        )

    def isin(self, test_elements):
        return LayerOperation(
            self,
            op.ISIN,
            window_op=WindowOperation.NONE,
            test_elements=test_elements,
        )

    def abs(self):
        return LayerOperation(
            self,
            op.ABS,
            window_op=WindowOperation.NONE,
        )

    def log(self):
        return LayerOperation(
            self,
            op.LOG,
            window_op=WindowOperation.NONE,
        )

    def log2(self):
        return LayerOperation(
            self,
            op.LOG2,
            window_op=WindowOperation.NONE,
        )

    def log10(self):
        return LayerOperation(
            self,
            op.LOG10,
            window_op=WindowOperation.NONE,
        )

    def exp(self):
        return LayerOperation(
            self,
            op.EXP,
            window_op=WindowOperation.NONE,
        )

    def exp2(self):
        return LayerOperation(
            self,
            op.EXP2,
            window_op=WindowOperation.NONE,
        )

    def clip(self, min=None, max=None): # pylint: disable=W0622
        # In the numpy 1 API np.clip(array) used a_max, a_min arguments and array.clip() used max and min as arguments
        # In numpy 2 they moved so that max and min worked on both, but still support a_max, and a_min on np.clip.
        # For now I'm only going to support the newer max/min everywhere notion, but I have to internally call
        # a_max, a_min so that yirgacheffe can work on older numpy installs.
        return LayerOperation(
            self,
            op.CLIP,
            window_op=WindowOperation.NONE,
            a_min=min,
            a_max=max,
        )

    def conv2d(self, weights):
        # A set of limitations that are just down to implementation time restrictions
        weights_width, weights_height = weights.shape
        if weights_width != weights_height:
            raise ValueError("Currently only square matrixes are supported for weights")
        padding = (weights_width - 1) / 2
        if padding != int(padding):
            raise ValueError("Currently weights dimensions must be odd")

        return LayerOperation(
            self,
            op.CONV2D,
            window_op=WindowOperation.NONE,
            buffer_padding=padding,
            weights=weights.astype(np.float32),
        )

    def numpy_apply(self, func, other=None):
        return LayerOperation(self, func, other)

    def shader_apply(self, func, other=None):
        return ShaderStyleOperation(self, func, other)

    def save(self, destination_layer, and_sum=False, callback=None, band=1):
        return LayerOperation(self).save(destination_layer, and_sum, callback, band)

    def parallel_save(self, destination_layer, and_sum=False, callback=None, parallelism=None, band=1):
        return LayerOperation(self).parallel_save(destination_layer, and_sum, callback, parallelism, band)

    def parallel_sum(self, callback=None, parallelism=None, band=1):
        return LayerOperation(self).parallel_sum(callback, parallelism, band)

    def sum(self):
        return LayerOperation(self).sum()

    def min(self):
        return LayerOperation(self).min()

    def max(self):
        return LayerOperation(self).max()


class LayerOperation(LayerMathMixin):

    @staticmethod
    def where(cond, a, b):
        return LayerOperation(
            cond,
            op.WHERE,
            rhs=a,
            other=b
        )

    @staticmethod
    def maximum(a, b):
        return LayerOperation(
            a,
            op.MAXIMUM,
            b,
            window_op=WindowOperation.UNION,
        )

    @staticmethod
    def minimum(a, b):
        return LayerOperation(
            a,
            op.MINIMUM,
            rhs=b,
            window_op=WindowOperation.UNION,
        )

    def __init__(
        self,
        lhs,
        operator=None,
        rhs=None,
        other=None,
        window_op=WindowOperation.NONE,
        buffer_padding=0,
        **kwargs
    ):
        self.ystep = constants.YSTEP
        self.kwargs = kwargs
        self.window_op = window_op
        self.buffer_padding = buffer_padding

        if lhs is None:
            raise ValueError("LHS on operation should not be none")
        self.lhs = lhs

        self.operator = operator

        if rhs is not None:
            if backend.isscalar(rhs):
                self.rhs = LayerConstant(rhs)
            elif isinstance(rhs, (backend.array_t)):
                if rhs.shape == ():
                    self.rhs = LayerConstant(rhs.item())
                else:
                    raise ValueError("Numpy arrays are no allowed")
            else:
                if not are_pixel_scales_equal_enough([lhs.pixel_scale, rhs.pixel_scale]):
                    raise ValueError("Not all layers are at the same pixel scale")
                self.rhs = rhs
        else:
            self.rhs = None

        if other is not None:
            if backend.isscalar(other):
                self.other = LayerConstant(other)
            elif isinstance(other, (backend.array_t)):
                if other.shape == ():
                    self.rhs = LayerConstant(other.item())
                else:
                    raise ValueError("Numpy arrays are no allowed")
            else:
                if not are_pixel_scales_equal_enough([lhs.pixel_scale, other.pixel_scale]):
                    raise ValueError("Not all layers are at the same pixel scale")
                self.other = other
        else:
            self.other = None

    def __str__(self):
        try:
            return f"({self.lhs} {self.operator} {self.rhs})"
        except AttributeError:
            try:
                return f"({self.operator} {self.lhs})"
            except AttributeError:
                return str(self.lhs)

    def __len__(self):
        return len(self.lhs)

    def __getstate__(self) -> object:
        odict = self.__dict__.copy()
        if isinstance(self.operator, types.LambdaType):
            odict['operator_dill'] = dumps(self.operator)
            del odict['operator']
        return odict

    def __setstate__(self, state):
        if 'operator_dill' in state:
            state['operator'] = loads(state['operator_dill'])
            del state['operator_dill']
        self.__dict__.update(state)

    @property
    def area(self) -> Area:
        # The type().__name__ here is to avoid a circular import dependancy
        lhs_area = self.lhs.area if not type(self.lhs).__name__ == "ConstantLayer" else None
        try:
            rhs_area = self.rhs.area if not type(self.rhs).__name__ == "ConstantLayer" else None
        except AttributeError:
            rhs_area = None
        try:
            other_area = self.other.area if not type(self.other).__name__ == "ConstantLayer" else None
        except AttributeError:
            other_area = None

        all_areas = []
        if lhs_area is not None:
            all_areas.append(lhs_area)
        if rhs_area is not None:
            all_areas.append(rhs_area)
        if other_area is not None:
            all_areas.append(other_area)

        match self.window_op:
            case WindowOperation.NONE:
                return all_areas[0]
            case WindowOperation.LEFT:
                return lhs_area
            case WindowOperation.RIGHT:
                assert rhs_area is not None
                return rhs_area
            case WindowOperation.INTERSECTION:
                intersection = Area(
                    left=max(x.left for x in all_areas),
                    top=min(x.top for x in all_areas),
                    right=min(x.right for x in all_areas),
                    bottom=max(x.bottom for x in all_areas)
                )
                if (intersection.left >= intersection.right) or (intersection.bottom >= intersection.top):
                    raise ValueError('No intersection possible')
                return intersection
            case WindowOperation.UNION:
                return Area(
                    left=min(x.left for x in all_areas),
                    top=max(x.top for x in all_areas),
                    right=max(x.right for x in all_areas),
                    bottom=min(x.bottom for x in all_areas)
                )

    @property
    def pixel_scale(self) -> PixelScale:
        # Because we test at construction that pixel scales for RHS/other are roughly equal,
        # I believe this should be sufficient...
        try:
            pixel_scale = self.lhs.pixel_scale
        except AttributeError:
            pixel_scale = None

        if pixel_scale is None:
            return self.rhs.pixel_scale
        return pixel_scale

    @property
    def window(self) -> Window:
        pixel_scale = self.pixel_scale
        area = self.area

        return Window(
            xoff=round_down_pixels(area.left / pixel_scale.xstep, pixel_scale.xstep),
            yoff=round_down_pixels(area.top / (pixel_scale.ystep * -1.0), pixel_scale.ystep * -1.0),
            xsize=round_up_pixels(
                (area.right - area.left) / pixel_scale.xstep, pixel_scale.xstep
            ),
            ysize=round_up_pixels(
                (area.top - area.bottom) / (pixel_scale.ystep * -1.0),
                (pixel_scale.ystep * -1.0)
            ),
        )

    @property
    def datatype(self):
        # TODO: Work out how to indicate type promotion via numpy
        return self.lhs.datatype

    @property
    def projection(self):
        try:
            return self.lhs.projection
        except AttributeError:
            return self.rhs.projection

    def _eval(self, area: Area, index: int, step: int, target_window:Optional[Window]=None):

        if self.buffer_padding:
            if target_window:
                target_window = target_window.grow(self.buffer_padding)
            pixel_scale = self.pixel_scale
            area = area.grow(self.buffer_padding * pixel_scale.xstep)
            # The index doesn't need updating because we updated area/window
            step += (2 * self.buffer_padding)

        lhs_data = self.lhs._eval(area, index, step, target_window)

        if self.operator is None:
            return lhs_data

        try:
            operator = backend.operator_map[self.operator]
        except KeyError:
            # Handles things like `numpy_apply` where a custom operator is provided
            operator = self.operator

        if self.other is not None:
            assert self.rhs is not None
            rhs_data = self.rhs._eval(area, index, step, target_window)
            other_data = self.other._eval(area, index, step, target_window)
            return operator(lhs_data, rhs_data, other_data, **self.kwargs)

        if self.rhs is not None:
            rhs_data = self.rhs._eval(area, index, step, target_window)
            return operator(lhs_data, rhs_data, **self.kwargs)

        return operator(lhs_data, **self.kwargs)

    def sum(self):
        # The result accumulator is float64, and for precision reasons
        # we force the sum to be done in float64 also. Otherwise we
        # see variable results depending on chunk size, as different parts
        # of the sum are done in different types.
        res = 0.0
        computation_window = self.window
        for yoffset in range(0, computation_window.ysize, self.ystep):
            step=self.ystep
            if yoffset+step > computation_window.ysize:
                step = computation_window.ysize - yoffset
            chunk = self._eval(self.area, yoffset, step, computation_window)
            res += backend.sum_op(chunk)
        return res

    def min(self):
        res = None
        computation_window = self.window
        for yoffset in range(0, computation_window.ysize, self.ystep):
            step=self.ystep
            if yoffset+step > computation_window.ysize:
                step = computation_window.ysize - yoffset
            chunk = self._eval(self.area, yoffset, step, computation_window)
            chunk_min = backend.min_op(chunk)
            if (res is None) or (res > chunk_min):
                res = chunk_min
        return res

    def max(self):
        res = None
        computation_window = self.window
        for yoffset in range(0, computation_window.ysize, self.ystep):
            step=self.ystep
            if yoffset+step > computation_window.ysize:
                step = computation_window.ysize - yoffset
            chunk = self._eval(self.area, yoffset, step, computation_window)
            chunk_max = backend.max_op(chunk)
            if (res is None) or (chunk_max > res):
                res = chunk_max
        return res

    def save(self, destination_layer, and_sum=False, callback=None, band=1):
        """
        Calling save will write the output of the operation to the provied layer.
        If you provide sum as true it will additionall compute the sum and return that.
        """

        if destination_layer is None:
            raise ValueError("Layer is required")
        try:
            band = destination_layer._dataset.GetRasterBand(band)
        except AttributeError as exc:
            raise ValueError("Layer must be a raster backed layer") from exc

        computation_window = self.window
        destination_window = destination_layer.window

        if (computation_window.xsize != destination_window.xsize) \
                or (computation_window.ysize != destination_window.ysize):
            raise ValueError("Destination raster window size does not match input raster window size.")

        total = 0.0

        for yoffset in range(0, computation_window.ysize, self.ystep):
            if callback:
                callback(yoffset / computation_window.ysize)
            step=self.ystep
            if yoffset+step > computation_window.ysize:
                step = computation_window.ysize - yoffset
            chunk = self._eval(self.area, yoffset, step, computation_window)
            if isinstance(chunk, (float, int)):
                chunk = backend.full((step, destination_window.xsize), chunk)
            band.WriteArray(
                backend.demote_array(chunk),
                destination_window.xoff,
                yoffset + destination_window.yoff,
            )
            if and_sum:
                total += backend.sum_op(chunk)
        if callback:
            callback(1.0)

        return total if and_sum else None

    def _parallel_worker(self, index, shared_mem, sem, np_dtype, width, input_queue, output_queue):
        arr = np.ndarray((self.ystep, width), dtype=np_dtype, buffer=shared_mem.buf)

        try:
            while True:
                # We acquire the lock so we know we have somewhere to put the
                # result before we take work. This is because in practice
                # it seems the writing to GeoTIFF is the bottleneck, and
                # we had workers taking a task, then waiting for somewhere to
                # write to for ages when other workers were exiting because there
                # was nothing to do.
                sem.acquire()

                task = input_queue.get()
                if task is None:
                    sem.release()
                    output_queue.put(None)
                    break
                yoffset, step = task

                result = self._eval(self.area, yoffset, step)
                backend.eval_op(result)

                arr[:step] = backend.demote_array(result)

                output_queue.put((index, yoffset, step))

        except Exception as e: # pylint: disable=W0718
            logger.exception(e)
            sem.release()
            output_queue.put(None)

    def _park(self):
        try:
            self.lhs._park()
        except AttributeError:
            pass
        try:
            self.rhs._park()
        except AttributeError:
            pass
        try:
            self.other._park()
        except AttributeError:
            pass

    def _parallel_save(self, destination_layer, and_sum=False, callback=None, parallelism=None, band=1):
        assert (destination_layer is not None) or and_sum
        computation_window = self.window
        if destination_layer is not None:
            try:
                band = destination_layer._dataset.GetRasterBand(band)
            except AttributeError as exc:
                raise ValueError("Layer must be a raster backed layer") from exc

            destination_window = destination_layer.window

            if (computation_window.xsize != destination_window.xsize) \
                    or (computation_window.ysize != destination_window.ysize):
                raise ValueError("Destination raster window size does not match input raster window size.")

            np_dtype = {
                gdal.GDT_Byte:    np.dtype('byte'),
                gdal.GDT_Float32: np.dtype('float32'),
                gdal.GDT_Float64: np.dtype('float64'),
                gdal.GDT_Int8:    np.dtype('int8'),
                gdal.GDT_Int16:   np.dtype('int16'),
                gdal.GDT_Int32:   np.dtype('int32'),
                gdal.GDT_Int64:   np.dtype('int64'),
                gdal.GDT_UInt16:  np.dtype('uint16'),
                gdal.GDT_UInt32:  np.dtype('uint32'),
                gdal.GDT_UInt64:  np.dtype('uint64'),
            }[band.DataType]
        else:
            band = None
            np_dtype = np.dtype('float64')

        # The parallel save will cause a fork on linux, so we need to
        # remove all SWIG references
        self._park()

        total = 0.0

        with multiprocessing.Manager() as manager:
            with SharedMemoryManager() as smm:

                worker_count = parallelism or multiprocessing.cpu_count()
                work_blocks = len(range(0, computation_window.ysize, self.ystep))
                worker_count = min(work_blocks, worker_count)

                mem_sem_cast = []
                for i in range(worker_count):
                    shared_buf = smm.SharedMemory(size=np_dtype.itemsize * self.ystep * computation_window.xsize)
                    cast_buf = np.ndarray((self.ystep, computation_window.xsize), dtype=np_dtype, buffer=shared_buf.buf)
                    cast_buf[:] = np.zeros((self.ystep, computation_window.xsize), np_dtype)
                    mem_sem_cast.append((shared_buf, Semaphore(), cast_buf))

                source_queue = manager.Queue()
                result_queue = manager.Queue()

                for yoffset in range(0, computation_window.ysize, self.ystep):
                    step = ((computation_window.ysize - yoffset)
                        if yoffset+self.ystep > computation_window.ysize
                        else self.ystep)
                    source_queue.put((
                        yoffset,
                        step
                    ))
                for _ in range(worker_count):
                    source_queue.put(None)

                if callback:
                    callback(0.0)

                workers = [Process(target=self._parallel_worker, args=(
                    i,
                    mem_sem_cast[i][0],
                    mem_sem_cast[i][1],
                    np_dtype,
                    computation_window.xsize,
                    source_queue,
                    result_queue
                )) for i in range(worker_count)]
                for worker in workers:
                    worker.start()

                sentinal_count = len(workers)
                retired_blocks = 0
                while sentinal_count > 0:
                    res = result_queue.get()
                    if res is None:
                        sentinal_count -= 1
                        continue
                    index, yoffset, step = res
                    _, sem, arr = mem_sem_cast[index]
                    if band:
                        band.WriteArray(
                            arr[0:step],
                            destination_window.xoff,
                            yoffset + destination_window.yoff,
                        )
                    if and_sum:
                        total += np.sum(np.array(arr[0:step]).astype(np.float64))
                    sem.release()
                    retired_blocks += 1
                    if callback:
                        callback(retired_blocks / work_blocks)

                processes = workers
                while processes:
                    candidates = [x for x in processes if not x.is_alive()]
                    for candidate in candidates:
                        candidate.join()
                        if candidate.exitcode:
                            for victim in processes:
                                victim.kill()
                            sys.exit(candidate.exitcode)
                        processes.remove(candidate)
                    time.sleep(0.01)

        return total if and_sum else None

    def parallel_save(self, destination_layer, and_sum=False, callback=None, parallelism=None, band=1):
        if destination_layer is None:
            raise ValueError("Layer is required")
        return self._parallel_save(destination_layer, and_sum, callback, parallelism, band)

    def parallel_sum(self, callback=None, parallelism=None, band=1):
        return self._parallel_save(None, True, callback, parallelism, band)

class ShaderStyleOperation(LayerOperation):

    def _eval(self, area, index, step, target_window=None):
        if target_window is None:
            target_window = self.window
        lhs_data = self.lhs._eval(area, index, step, target_window)
        if self.rhs is not None:
            rhs_data = self.rhs._eval(area, index, step, target_window)
        else:
            rhs_data = None

        # Constant results make this a bit messier. Might in future
        # be nicer to promote them to arrays sooner?
        if isinstance(lhs_data, (int, float)):
            if rhs_data is None:
                return self.operator(lhs_data, **self.kwargs)
            if isinstance(rhs_data, (int, float)):
                return self.operator(lhs_data, rhs_data, **self.kwargs)
            else:
                result = np.empty_like(rhs_data)
        else:
            result = np.empty_like(lhs_data)

        window = self.window
        for yoffset in range(step):
            for xoffset in range(window.xsize):
                try:
                    lhs_val = lhs_data[yoffset][xoffset]
                except TypeError:
                    lhs_val = lhs_data
                if rhs_data is not None:
                    try:
                        rhs_val = rhs_data[yoffset][xoffset]
                    except TypeError:
                        rhs_val = rhs_data
                    result[yoffset][xoffset] = self.operator(lhs_val, rhs_val, **self.kwargs)
                else:
                    result[yoffset][xoffset] = self.operator(lhs_val, **self.kwargs)

        return result

# We provide these module level accessors as it's often nicer to write `log(x/y)` rather than `(x/y).log()`
where = LayerOperation.where
minumum = LayerOperation.minimum
maximum = LayerOperation.maximum
clip = LayerOperation.clip
log = LayerOperation.log
log2 = LayerOperation.log2
log10 = LayerOperation.log10
exp = LayerOperation.exp
exp2 = LayerOperation.exp2
nan_to_num = LayerOperation.nan_to_num
isin = LayerOperation.isin
abs = LayerOperation.abs # pylint: disable=W0622



import os
import tempfile
from math import ceil, floor

import pytest

from helpers import gdal_dataset_of_region
from yirgacheffe.layers import UniformAreaLayer
from yirgacheffe.window import Area, Window

# A UniformAreaLayer is a hack for the fact that we have these raster layers
# that represent the area of each pixel in WSG84, and they're all the same
# value across the layer, but the GeoTIFFs are super slow to unpack. Thus the
# work around is to make these GeoTIFFS that are just one column of pixels,
# and this layer pretends that it's a 360 degree wide layer.

@pytest.mark.parametrize(
    "pixel_scale",
    [
        1.0,
        0.2345678,
    ]
)
def test_open_uniform_area_layer(pixel_scale: float) -> None:
    with tempfile.TemporaryDirectory() as tempdir:
        path = os.path.join(tempdir, "test.tif")
        area = Area(
            floor(-180 / pixel_scale) * pixel_scale,
            ceil(90 / pixel_scale) * pixel_scale,
            (floor(-180 / pixel_scale) * pixel_scale) + pixel_scale,
            floor(-90 / pixel_scale) * pixel_scale
        )
        dataset = gdal_dataset_of_region(area, pixel_scale, filename=path)
        assert dataset.RasterXSize == 1
        assert dataset.RasterYSize == ceil(180 / pixel_scale)
        dataset.Close()

        layer = UniformAreaLayer.layer_from_file(path)
        assert layer.pixel_scale == (pixel_scale, -pixel_scale)
        assert layer.area == Area(
            floor(-180 / pixel_scale) * pixel_scale,
            ceil(90 / pixel_scale) * pixel_scale,
            ceil(180 / pixel_scale) * pixel_scale,
            floor(-90 / pixel_scale) * pixel_scale
        )
        assert layer.window == Window(
            0,
            0,
            ceil((layer.area.right - layer.area.left) / pixel_scale),
            ceil((layer.area.top - layer.area.bottom) / pixel_scale)
        )

def test_incorrect_tiff_for_uniform_area() -> None:
    area = Area(-10, 10, 10, -10)
    dataset = gdal_dataset_of_region(area, 1.0)
    with pytest.raises(ValueError):
        _ = UniformAreaLayer(dataset)

def test_set_intersection() -> None:
    with tempfile.TemporaryDirectory() as tempdir:
        path = os.path.join(tempdir, "test.tif")
        area = Area(-180, 90, -179, -90)
        dataset = gdal_dataset_of_region(area, 1.0, filename=path)
        assert dataset.RasterXSize == 1
        assert dataset.RasterYSize == 180
        dataset.Close()

        layer = UniformAreaLayer.layer_from_file(path)
        layer.set_window_for_intersection(Area(-10, 10, 10, -10))
        assert layer.area == Area(-10, 10, 10, -10)
        assert layer.window == Window(170, 80, 20, 20)

        layer.reset_window()
        assert layer.area == Area(-180, 90, 180, -90)
        assert layer.window == Window(0, 0, 360, 180)



from enum import Enum

class operators(Enum):
    ADD = 1
    SUB = 2
    MUL = 3
    TRUEDIV = 4
    POW = 5
    EQ = 6
    NE = 7
    LT = 8
    LE = 9
    GT = 10
    GE = 11
    AND = 12
    OR = 13
    LOG = 14
    LOG2 = 15
    LOG10 = 16
    EXP = 17
    EXP2 = 18
    CLIP = 19
    WHERE = 20
    MIN = 21
    MAX = 22
    SUM = 23
    MINIMUM = 24
    MAXIMUM = 25
    NAN_TO_NUM = 26
    ISIN = 27
    REMAINDER = 28
    FLOORDIV = 29
    CONV2D = 30
    ABS = 31



import math
from typing import Tuple

import pytest

from yirgacheffe import WGS_84_PROJECTION
from yirgacheffe.layers import YirgacheffeLayer
from yirgacheffe.window import Area, PixelScale

def test_pixel_to_latlng_unsupported_projection() -> None:
    layer = YirgacheffeLayer(
        Area(-10, 10, 10, -10),
        PixelScale(0.02, -0.02),
        "OTHER PROJECTION"
    )
    with pytest.raises(NotImplementedError):
        _ = layer.latlng_for_pixel(10, 10)

def test_pixel_from_latlng_unsupported_projection() -> None:
    layer = YirgacheffeLayer(
        Area(-10, 10, 10, -10),
        PixelScale(0.02, -0.02),
        "OTHER PROJECTION"
    )
    with pytest.raises(NotImplementedError):
        _ = layer.pixel_for_latlng(10.0, 10.0)

@pytest.mark.parametrize(
    "area,pixel,expected",
    [
        (
            Area(-10, 10, 10, -10),
            (0, 0),
            (10.0, -10.0)
        ),
        (
            Area(-10, 10, 10, -10),
            (1, 1),
            (9.8, -9.8)
        ),
        (
            Area(-10, 10, 10, -10),
            (101, 101),
            (-10.2, 10.2)
        ),
        (
            Area(-10, 10, 10, -10),
            (-1, -1),
            (10.2, -10.2)
        ),
        (
            Area(10, 10, 20, -10),
            (1, 1),
            (9.8, 10.2)
        ),
        (
            Area(-10, -10, 10, -20),
            (1, 1),
            (-10.2, -9.8)
        ),
    ]
)
def test_latlng_for_pixel(area: Area, pixel: Tuple[int,int], expected: Tuple[float,float]) -> None:
    layer = YirgacheffeLayer(
        area,
        PixelScale(0.2, -0.2),
        WGS_84_PROJECTION
    )
    result = layer.latlng_for_pixel(*pixel)
    assert math.isclose(result[0], expected[0])
    assert math.isclose(result[1], expected[1])

@pytest.mark.parametrize(
    "area,coord,expected",
    [
        (
            Area(-10, 10, 10, -10),
            (10.0, -10.0),
            (0, 0)
        ),
        (
            Area(-10, 10, 10, -10),
            (9.8, -9.8),
            (1, 1)
        ),
        (
            Area(-10, 10, 10, -10),
            (0.0, 0.0),
            (50, 50)
        ),
    ]
)
def test_pixel_for_latlng(area: Area, coord: Tuple[float,float], expected: Tuple[int,int]) -> None:
    layer = YirgacheffeLayer(
        area,
        PixelScale(0.2, -0.2),
        WGS_84_PROJECTION
    )
    result = layer.pixel_for_latlng(*coord)
    assert result == expected


@pytest.mark.parametrize(
    "area,window,pixel,expected",
    [
        (
            Area(-10, 10, 10, -10),
            Area(-5, 5, 5, -5),
            (0, 0),
            (5.0, -5.0)
        ),
        (
            Area(-10, 10, 10, -10),
            Area(-5, 5, 5, -5),
            (1, 1),
            (4.8, -4.8)
        ),
        (
            Area(-10, 10, 10, -10),
            Area(-5, 5, 5, -5),
            (101, 101),
            (-15.2, 15.2)
        ),
        (
            Area(-10, 10, 10, -10),
            Area(-5, 5, 5, -5),
            (-1, -1),
            (5.2, -5.2)
        ),
        (
            Area(10, 10, 20, -10),
            Area(15, 5, 20, -5),
            (1, 1),
            (4.8, 15.2)
        ),
        (
            Area(-10, -10, 10, -20),
            Area(-5, -15, 5, -20),
            (1, 1),
            (-15.2, -4.8)
        ),
    ]
)
def test_latlng_for_pixel_with_intersection(
    area: Area,
    window: Area,
    pixel: Tuple[int,int],
    expected: Tuple[float,float]
) -> None:
    layer = YirgacheffeLayer(
        area,
        PixelScale(0.2, -0.2),
        WGS_84_PROJECTION
    )
    layer.set_window_for_intersection(window)
    result = layer.latlng_for_pixel(*pixel)
    assert math.isclose(result[0], expected[0])
    assert math.isclose(result[1], expected[1])

@pytest.mark.parametrize(
    "area,window,coord,expected",
    [
        (
            Area(-10, 10, 10, -10),
            Area(-5, 5, 5, -5),
            (5.0, -5.0),
            (0, 0)
        ),
        (
            Area(-10, 10, 10, -10),
            Area(-5, 5, 5, -5),
            (4.8, -4.8),
            (1, 1)
        ),
        (
            Area(-10, 10, 10, -10),
            Area(-5, 5, 5, -5),
            (0.0, 0.0),
            (25, 25)
        ),
    ]
)
def test_pixel_for_latlng_with_intersection(
    area: Area,
    window: Area,
    coord: Tuple[float,float],
    expected: Tuple[int,int]
) -> None:
    layer = YirgacheffeLayer(
        area,
        PixelScale(0.2, -0.2),
        WGS_84_PROJECTION
    )
    layer.set_window_for_intersection(window)
    result = layer.pixel_for_latlng(*coord)
    assert result == expected



import math
import os
from typing import Any, Optional, TypeVar, Union

import numpy as np
from osgeo import gdal

from .. import WGS_84_PROJECTION
from ..window import Area, PixelScale, Window
from ..rounding import round_up_pixels
from .base import YirgacheffeLayer
from ..backends import backend

# Still to early to require Python 3.11 :/
RasterLayerT = TypeVar("RasterLayerT", bound="RasterLayer")

class InvalidRasterBand(Exception):
    def __init__ (self, band):
        self.band = band

class RasterLayer(YirgacheffeLayer):
    """Layer provides a wrapper around a gdal dataset/band that also records offset state so that
    we can work with maps over different geographic regions but work withing a particular frame
    of reference."""

    @staticmethod
    def empty_raster_layer(
        area: Area,
        scale: PixelScale,
        datatype: int,
        filename: Optional[str]=None,
        projection: str=WGS_84_PROJECTION,
        name: Optional[str]=None,
        compress: bool=True,
        nodata: Optional[Union[float,int]]=None,
        nbits: Optional[int]=None,
        threads: Optional[int]=None,
        bands: int=1
    ) -> RasterLayerT:
        abs_xstep, abs_ystep = abs(scale.xstep), abs(scale.ystep)

        # We treat the provided area as aspirational, and we need to align it to pixel boundaries
        pixel_friendly_area = Area(
            left=math.floor(area.left / abs_xstep) * abs_xstep,
            right=math.ceil(area.right / abs_xstep) * abs_xstep,
            top=math.ceil(area.top / abs_ystep) * abs_ystep,
            bottom=math.floor(area.bottom / abs_ystep) * abs_ystep,
        )

        options = []
        if threads is not None:
            options.append(f"NUM_THREADS={threads}")
        if nbits is not None:
            options.append(f"NBITS={nbits}")

        if filename:
            driver = gdal.GetDriverByName('GTiff')
            options.append('BIGTIFF=YES')
            if compress:
                options.append('COMPRESS=LZW')
            else:
                options.append('COMPRESS=NONE')
        else:
            driver = gdal.GetDriverByName('mem')
            filename = 'mem'
            compress = False
        dataset = driver.Create(
            filename,
            round_up_pixels((pixel_friendly_area.right - pixel_friendly_area.left) / abs_xstep, abs_xstep),
            round_up_pixels((pixel_friendly_area.top - pixel_friendly_area.bottom) / abs_ystep, abs_ystep),
            bands,
            datatype,
            options
        )
        dataset.SetGeoTransform([
            pixel_friendly_area.left, scale.xstep, 0.0, pixel_friendly_area.top, 0.0, scale.ystep
        ])
        dataset.SetProjection(projection)
        if nodata is not None:
            dataset.GetRasterBand(1).SetNoDataValue(nodata)
        return RasterLayer(dataset, name=name)

    @staticmethod
    def empty_raster_layer_like(
        layer: Any,
        filename: Optional[str]=None,
        area: Optional[Area]=None,
        datatype: Optional[int]=None,
        compress: bool=True,
        nodata: Optional[Union[float,int]]=None,
        nbits: Optional[int]=None,
        threads: Optional[int]=None,
        bands: int=1
    ) -> RasterLayerT:
        width = layer.window.xsize
        height = layer.window.ysize
        if area is None:
            area = layer.area
        assert area is not None

        scale = layer.pixel_scale
        if scale is None:
            raise ValueError("Can not work out area without explicit pixel scale")
        abs_xstep, abs_ystep = abs(scale.xstep), abs(scale.ystep)
        width = round_up_pixels((area.right - area.left) / abs_xstep, abs_xstep)
        height = round_up_pixels((area.top - area.bottom) / abs_ystep, abs_ystep)
        geo_transform = (
            area.left, scale.xstep, 0.0, area.top, 0.0, scale.ystep
        )

        options = []
        if threads is not None:
            options.append(f"NUM_THREADS={threads}")
        if nbits is not None:
            options.append(f"NBITS={nbits}")

        if filename:
            driver = gdal.GetDriverByName('GTiff')
            options.append('BIGTIFF=YES')
            if compress:
                options.append('COMPRESS=LZW')
            else:
                options.append('COMPRESS=NONE')
        else:
            driver = gdal.GetDriverByName('mem')
            filename = 'mem'
            compress = False
        dataset = driver.Create(
            filename,
            width,
            height,
            bands,
            datatype if datatype is not None else layer.datatype,
            options,
        )
        dataset.SetGeoTransform(geo_transform)
        dataset.SetProjection(layer.projection)
        if nodata is not None:
            dataset.GetRasterBand(1).SetNoDataValue(nodata)

        return RasterLayer(dataset)

    @classmethod
    def scaled_raster_from_raster(
        cls,
        source: RasterLayerT,
        new_pixel_scale: PixelScale,
        filename: Optional[str]=None,
        compress: bool=True,
        algorithm: int=gdal.GRA_NearestNeighbour,
    ) -> RasterLayerT:
        source_dataset = source._dataset
        old_pixel_scale = source.pixel_scale
        assert old_pixel_scale

        x_scale = old_pixel_scale.xstep / new_pixel_scale.xstep
        y_scale = old_pixel_scale.ystep / new_pixel_scale.ystep
        new_width = round_up_pixels(source_dataset.RasterXSize * x_scale,
            abs(new_pixel_scale.xstep))
        new_height = round_up_pixels(source_dataset.RasterYSize * y_scale,
            abs(new_pixel_scale.ystep))

        # in yirgacheffe we like to have things aligned to the pixel_scale, so work
        # out new top left corner
        new_left = math.floor((source.area.left / new_pixel_scale.xstep)) * new_pixel_scale.xstep
        new_top = math.ceil((source.area.top / new_pixel_scale.ystep)) * new_pixel_scale.ystep

        # now build a target dataset
        options = []
        if filename:
            driver = gdal.GetDriverByName('GTiff')
            options.append('BIGTIFF=YES')
            if compress:
                options.append('COMPRESS=LZW')
        else:
            driver = gdal.GetDriverByName('mem')
            filename = 'mem'
            compress = False
        dataset = driver.Create(
            filename,
            new_width,
            new_height,
            1,
            source.datatype,
            options
        )
        dataset.SetGeoTransform((
            new_left, new_pixel_scale.xstep, 0.0,
            new_top, 0.0, new_pixel_scale.ystep
        ))
        dataset.SetProjection(source_dataset.GetProjection())

        # now use gdal to do the reprojection
        gdal.ReprojectImage(source_dataset, dataset, eResampleAlg=algorithm)

        return RasterLayer(dataset)

    @classmethod
    def layer_from_file(cls, filename: str, band: int = 1) -> RasterLayerT:
        try:
            dataset = gdal.Open(filename, gdal.GA_ReadOnly)
        except RuntimeError as exc:
            # With exceptions on GDAL now returns the wrong (IMHO) exception
            raise FileNotFoundError(filename) from exc
        try:
            _ = dataset.GetRasterBand(band)
        except RuntimeError as exc:
            raise InvalidRasterBand(band) from exc
        return cls(dataset, filename, band)

    def __init__(self, dataset: gdal.Dataset, name: Optional[str] = None, band: int = 1):
        if not dataset:
            raise ValueError("None is not a valid dataset")

        transform = dataset.GetGeoTransform()
        scale = PixelScale(transform[1], transform[5])
        area = Area(
            left=transform[0],
            top=transform[3],
            right=transform[0] + (dataset.RasterXSize * scale.xstep),
            bottom=transform[3] + (dataset.RasterYSize * scale.ystep),
        )

        super().__init__(
            area,
            scale,
            dataset.GetProjection(),
            name=name
        )

        # The constructor works out the window from the area
        # so sanity check that the calculated window matches the
        # dataset's dimensions
        assert self.window == Window(0, 0, dataset.RasterXSize, dataset.RasterYSize)

        self._dataset = dataset
        self._dataset_path = dataset.GetDescription()
        self._band = band
        self._raster_xsize = dataset.RasterXSize
        self._raster_ysize = dataset.RasterYSize

    def close(self):
        try:
            if self._dataset:
                try:
                    self._dataset.Close()
                except AttributeError:
                    pass
                del self._dataset
        except AttributeError:
            # Don't error if close was already called
            pass

    def __getstate__(self) -> object:
        # Only support pickling on file backed layers (ideally read only ones...)
        if not os.path.isfile(self._dataset_path):
            raise ValueError("Can not pickle layer that is not file backed.")
        odict = self.__dict__.copy()
        self._park()
        del odict['_dataset']
        return odict

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._unpark()

    def _park(self):
        try:
            self._dataset.Close()
        except AttributeError:
            pass
        self._dataset = None

    def _unpark(self):
        if getattr(self, "_dataset", None) is None:
            try:
                self._dataset = gdal.Open(self._dataset_path)
            except RuntimeError as exc:
                raise FileNotFoundError(f"Failed to open pickled raster {self._dataset_path}") from exc

    @property
    def datatype(self) -> int:
        if self._dataset is None:
            self._unpark()
        assert self._dataset
        return self._dataset.GetRasterBand(1).DataType

    def read_array_with_window(self, xoffset: int, yoffset: int, xsize: int, ysize: int, window: Window) -> Any:
        if self._dataset is None:
            self._unpark()
        assert self._dataset

        if (xsize <= 0) or (ysize <= 0):
            raise ValueError("Request dimensions must be positive and non-zero")

        # if we're dealing with an intersection, we can just read the data directly,
        # otherwise we need to read the data into another array with suitable padding
        target_window = Window(
            window.xoff + xoffset,
            window.yoff + yoffset,
            xsize,
            ysize
        )
        source_window = Window(
            xoff=0,
            yoff=0,
            xsize=self._raster_xsize,
            ysize=self._raster_ysize,
        )
        try:
            intersection = Window.find_intersection([source_window, target_window])
        except ValueError:
            return backend.zeros((ysize, xsize))

        if target_window == intersection:
            # The target window is a subset of or equal to the source, so we can just ask for the data
            data = backend.promote(self._dataset.GetRasterBand(self._band).ReadAsArray(*intersection.as_array_args))
            return data
        else:
            # We should read the intersection from the array, and the rest should be zeros
            subset = backend.promote(self._dataset.GetRasterBand(self._band).ReadAsArray(*intersection.as_array_args))
            region = np.array((
                (
                    (intersection.yoff - window.yoff) - yoffset,
                    (ysize - ((intersection.yoff - window.yoff) + intersection.ysize)) + yoffset,
                ),
                (
                    (intersection.xoff - window.xoff) - xoffset,
                    xsize - ((intersection.xoff - window.xoff) + intersection.xsize) + xoffset,
                )
            )).astype(int)
            data = backend.pad(subset, region, mode='constant')
            return data



import h3
import numpy as np
import pytest
from osgeo import gdal

from yirgacheffe import WGS_84_PROJECTION
from yirgacheffe.layers import RasterLayer, H3CellLayer
from yirgacheffe.window import Area, PixelScale
from yirgacheffe.operators import ShaderStyleOperation
from yirgacheffe.backends import backend


@pytest.mark.parametrize(
    "cell_id,is_valid,expected_zoom",
    [
        ("hello", False, 0),
        ("88972eac11fffff", True, 8),
    ]
)
def test_h3_layer(cell_id: str, is_valid: bool, expected_zoom: int) -> None:
    if is_valid:
        with H3CellLayer(cell_id, PixelScale(0.001, -0.001), WGS_84_PROJECTION) as layer:
            assert layer.zoom == expected_zoom
            assert layer.projection == WGS_84_PROJECTION

            # without getting too deep, we'd expect a mix of zeros and ones in the data
            window = layer.window
            one_count = 0
            for yoffset in range(window.ysize):
                data = layer.read_array(0, yoffset, window.xsize, 1)
                assert data.shape == (1, window.xsize)
                one_count += data.sum()
            assert one_count != 0
    else:
        with pytest.raises(ValueError):
            with H3CellLayer(cell_id, PixelScale(0.001, -0.001), WGS_84_PROJECTION) as _layer:
                pass

@pytest.mark.parametrize(
    "lat,lng",
    [
        (0.0, 0.0),
        (0.0, 45.0),
        (45.0, 0.0),
        (45.0, 45.0),
        (85.0, 0.0),
        (85.0, 45.0),
        (1.0, 95.0),
    ]
)
def test_h3_layer_magnifications(lat: float, lng: float) -> None:
    for zoom in range(6, 10):
        cell_id = h3.latlng_to_cell(lat, lng, zoom)
        h3_layer = H3CellLayer(cell_id, PixelScale(0.000898315284120,-0.000898315284120), WGS_84_PROJECTION)

        on_cell_count = h3_layer.sum()
        total_count = h3_layer.window.xsize * h3_layer.window.ysize
        assert 0 < on_cell_count < total_count

@pytest.mark.parametrize(
    "lat,lng",
    [
        (0.0, 0.0),
        (0.0, 45.0),
        (45.0, 0.0),
        (45.0, 45.0),
        (85.0, 0.0),
        (85.0, 45.0),
        (1.0, 95.0),
    ]
)
def test_h3_layer_not_clipped(lat: float, lng: float) -> None:
    for zoom in range(6, 10):
        cell_id = h3.latlng_to_cell(lat, lng, zoom)
        scale = PixelScale(0.000898315284120,-0.000898315284120)
        h3_layer = H3CellLayer(cell_id, scale, WGS_84_PROJECTION)

        on_cell_count = h3_layer.sum()
        assert on_cell_count > 0.0

        before_window = h3_layer.window
        abs_xstep, abs_ystep = abs(scale.xstep), abs(scale.ystep)
        expanded_area = Area(
            left=h3_layer.area.left - (12 * abs_xstep),
            top=h3_layer.area.top + (12 * abs_ystep),
            right=h3_layer.area.right + (12 * abs_xstep),
            bottom=h3_layer.area.bottom - (12 * abs_ystep)
        )
        h3_layer.set_window_for_union(expanded_area)
        assert h3_layer.window > before_window

        expanded_on_cell_count = h3_layer.sum()
        assert expanded_on_cell_count == on_cell_count

        # whilst we're here, check that we do have an empty border (i.e., does window for union do the right thing)
        assert np.sum(h3_layer.read_array(0, 0, h3_layer.window.xsize, 2)) == 0.0
        assert np.sum(h3_layer.read_array(0, h3_layer.window.ysize - 2, h3_layer.window.xsize, 2)) == 0.0
        assert np.sum(h3_layer.read_array(0, 0, 2, h3_layer.window.ysize)) == 0.0
        assert np.sum(h3_layer.read_array(h3_layer.window.xsize - 2, 0, 2, h3_layer.window.ysize)) == 0.0

@pytest.mark.parametrize(
    "lat,lng",
    [
        (0.0, 0.0),
        (0.0, 45.0),
        (45.0, 0.0),
        (45.0, 45.0),
        (85.0, 0.0),
        (85.0, 45.0),
        (1.0, 95.0),
    ]
)
def test_h3_layer_clipped(lat: float, lng: float) -> None:
    for zoom in range(6, 8):
        cell_id = h3.latlng_to_cell(lat, lng, zoom)
        scale = PixelScale(0.000898315284120,-0.000898315284120)
        h3_layer = H3CellLayer(cell_id, scale, WGS_84_PROJECTION)

        on_cell_count = h3_layer.sum()
        assert on_cell_count > 0.0

        before_window = h3_layer.window
        abs_xstep, abs_ystep = abs(scale.xstep), abs(scale.ystep)
        expanded_area = Area(
            left=h3_layer.area.left + (2 * abs_xstep),
            top=h3_layer.area.top - (2 * abs_ystep),
            right=h3_layer.area.right - (2 * abs_xstep),
            bottom=h3_layer.area.bottom + (2 * abs_ystep)
        )
        h3_layer.set_window_for_intersection(expanded_area)
        assert h3_layer.window < before_window

        shrunk_on_cell_count = h3_layer.sum()
        assert shrunk_on_cell_count < on_cell_count

        # whilst we're here, check that we do have an filled border (i.e., does window for
        # intersection do the right thing)
        assert np.sum(backend.demote_array(h3_layer.read_array(0, 0, h3_layer.window.xsize, 5))) > 0.0
        assert np.sum(backend.demote_array(h3_layer.read_array(0, h3_layer.window.ysize - 5, h3_layer.window.xsize, 5))) > 0.0
        assert np.sum(backend.demote_array(h3_layer.read_array(0, 0, 5, h3_layer.window.ysize))) > 0.0
        assert np.sum(backend.demote_array(h3_layer.read_array(h3_layer.window.xsize - 5, 0, 5, h3_layer.window.ysize))) > 0.0

@pytest.mark.parametrize(
    "lat,lng",
    [
        (50.0, 179.9),
        (50.0, -179.9),
    ]
)
def test_h3_layer_wrapped_on_projection(lat: float, lng: float) -> None:
    cell_id = h3.latlng_to_cell(lat, lng, 3)
    scale = PixelScale(0.01, -0.01)
    h3_layer = H3CellLayer(cell_id, scale, WGS_84_PROJECTION)

    # Just sanity check this test has caught a cell that wraps the entire planet and is testing
    # what we think it is testing:
    assert h3_layer.window.xsize == (360 / 0.01)

    area = h3_layer.sum()
    assert area > 0.0 # sanity check

    # Go around the cell neighbours, of which some will not wrap the planet, and
    # check they are all of a similarish size - we had a bug early on where we'd
    # mistakenly invert the area for the band, counting all the cells across the planet
    for cell_id in h3.grid_ring(cell_id, 1):
        neighbour = H3CellLayer(cell_id, scale, WGS_84_PROJECTION)
        neighbour_area = neighbour.sum()
        # We're happy if they're within 10% for now
        assert abs((neighbour_area - area) / area) < 0.1


    before_window = h3_layer.window
    _, abs_ystep = abs(scale.xstep), abs(scale.ystep)
    expanded_area = Area(
        left=h3_layer.area.left,
        top=h3_layer.area.top + (22 * abs_ystep),
        right=h3_layer.area.right,
        bottom=h3_layer.area.bottom - (22 * abs_ystep)
    )
    h3_layer.set_window_for_union(expanded_area)
    assert h3_layer.window.xsize == before_window.xsize
    assert h3_layer.window.xoff == 0
    assert before_window.xoff == 0
    assert h3_layer.window.yoff < 0
    assert before_window.yoff == 0
    assert h3_layer.window.ysize > before_window.ysize

    expanded_area = h3_layer.sum()
    assert expanded_area == area

    # whilst we're here, check that we do have an empty border (i.e., does window for union do the right thing)
    assert np.sum(backend.demote_array(h3_layer.read_array(0, 0, h3_layer.window.xsize, 2))) == 0.0
    assert np.sum(backend.demote_array(h3_layer.read_array(0, h3_layer.window.ysize - 2, h3_layer.window.xsize, 2))) == 0.0

def test_h3_layer_overlapped():
    # This is based on a regression, where somehow I made tiles not tesselate properly
    left, top = (121.26706, 19.45338)
    right, bottom = (121.62494, 19.18478)
    scale = PixelScale(0.000898315284120,-0.000898315284120)

    cells = h3.geo_to_cells(h3.LatLngPoly([
        (top, left),
        (top, right),
        (bottom, right),
        (bottom, left),
    ],), 7)

    tiles = [
        H3CellLayer(cell_id, scale, WGS_84_PROJECTION)
    for cell_id in cells]

    union = RasterLayer.find_union(tiles)
    union = union.grow(0.02)

    scratch = RasterLayer.empty_raster_layer(union, scale, gdal.GDT_Float64)

    # In scratch we should only have 0 or 1 values, but if there are any overlaps we should get 2s...

    for tile in tiles:
        scratch.reset_window()
        layers = [scratch, tile]
        intersection = RasterLayer.find_intersection(layers)
        for layer in layers:
            layer.set_window_for_intersection(intersection)
        result = scratch + tile
        result.save(scratch)

    scratch.reset_window()
    calc = scratch > 1.5
    assert calc.sum() == 0



import os
import tempfile

import numpy as np
import pytest
import torch

import yirgacheffe
from helpers import gdal_dataset_with_data
from yirgacheffe.layers import RasterLayer, ConstantLayer
from yirgacheffe.operators import LayerOperation

# These tests are marked skip for MLX, because there seems to be a problem with
# calling mx.eval in the tests for parallel save on Linux (which is what we use
# for github actions for instance). They do pass on macOS - but that could also
# just be down to Python versions. To add further complication, if I just run
# this file along on linux with MLX it passes fine 🤦
#
# It seems that under the hood MLX is doing some threading of its own and my
# guess is that that's interacting with the Python threading here.


@pytest.mark.skipif(yirgacheffe.backends.BACKEND != "NUMPY", reason="Only applies for numpy")
def test_add_byte_layers() -> None:
    with tempfile.TemporaryDirectory() as tempdir:
        path1 = os.path.join(tempdir, "test1.tif")
        data1 = np.array([[1, 2, 3, 4], [5, 6, 7, 8]])
        dataset1 = gdal_dataset_with_data((0.0, 0.0), 0.02, data1, filename=path1)
        dataset1.Close()

        path2 = os.path.join(tempdir, "test2.tif")
        data2 = np.array([[10, 20, 30, 40], [50, 60, 70, 80]])
        dataset2 = gdal_dataset_with_data((0.0, 0.0), 0.02, data2, filename=path2)
        dataset2.Close()

        layer1 = RasterLayer.layer_from_file(path1)
        layer2 = RasterLayer.layer_from_file(path2)
        result = RasterLayer.empty_raster_layer_like(layer1)

        comp = layer1 + layer2
        comp.parallel_save(result)

        expected = data1 + data2
        actual = result.read_array(0, 0, 4, 2)

        assert (expected == actual).all()

@pytest.mark.skipif(yirgacheffe.backends.BACKEND != "NUMPY", reason="Only applies for numpy")
def test_add_byte_layers_and_sum() -> None:
    with tempfile.TemporaryDirectory() as tempdir:
        path1 = os.path.join(tempdir, "test1.tif")
        data1 = np.array([[1, 2, 3, 4], [5, 6, 7, 8]])
        dataset1 = gdal_dataset_with_data((0.0, 0.0), 0.02, data1, filename=path1)
        dataset1.Close()

        path2 = os.path.join(tempdir, "test2.tif")
        data2 = np.array([[10, 20, 30, 40], [50, 60, 70, 80]])
        dataset2 = gdal_dataset_with_data((0.0, 0.0), 0.02, data2, filename=path2)
        dataset2.Close()

        layer1 = RasterLayer.layer_from_file(path1)
        layer2 = RasterLayer.layer_from_file(path2)
        result = RasterLayer.empty_raster_layer_like(layer1)

        comp = layer1 + layer2
        sum_total = comp.parallel_save(result, and_sum=True)

        expected = data1 + data2
        actual = result.read_array(0, 0, 4, 2)

        assert (expected == actual).all()
        assert sum_total == expected.sum()

@pytest.mark.skipif(yirgacheffe.backends.BACKEND != "NUMPY", reason="Only applies for numpy")
def test_parallel_sum() -> None:
    with tempfile.TemporaryDirectory() as tempdir:
        path1 = os.path.join(tempdir, "test1.tif")
        data1 = np.array([[1, 2, 3, 4], [5, 6, 7, 8]])
        dataset1 = gdal_dataset_with_data((0.0, 0.0), 0.02, data1, filename=path1)
        dataset1.Close()

        path2 = os.path.join(tempdir, "test2.tif")
        data2 = np.array([[10, 20, 30, 40], [50, 60, 70, 80]])
        dataset2 = gdal_dataset_with_data((0.0, 0.0), 0.02, data2, filename=path2)
        dataset2.Close()

        layer1 = RasterLayer.layer_from_file(path1)
        layer2 = RasterLayer.layer_from_file(path2)

        comp = layer1 + layer2
        sum_total = comp.parallel_sum()

        expected = data1 + data2
        assert sum_total == expected.sum()

@pytest.mark.skipif(yirgacheffe.backends.BACKEND != "NUMPY", reason="Only applies for numpy")
@pytest.mark.parametrize("skip,expected_steps", [
    (1, [0.0, 0.5, 1.0]),
    (2, [0.0, 1.0]),
])
def test_parallel_with_different_skip(skip, expected_steps) -> None:
    with tempfile.TemporaryDirectory() as tempdir:
        path1 = os.path.join(tempdir, "test1.tif")
        data1 = np.array([[1, 2, 3, 4], [5, 6, 7, 8]])
        dataset1 = gdal_dataset_with_data((0.0, 0.0), 0.02, data1, filename=path1)
        dataset1.Close()

        path2 = os.path.join(tempdir, "test2.tif")
        data2 = np.array([[10, 20, 30, 40], [50, 60, 70, 80]])
        dataset2 = gdal_dataset_with_data((0.0, 0.0), 0.02, data2, filename=path2)
        dataset2.Close()

        layer1 = RasterLayer.layer_from_file(path1)
        layer2 = RasterLayer.layer_from_file(path2)
        result = RasterLayer.empty_raster_layer_like(layer1)

        callback_possitions = []

        comp = layer1 + layer2
        comp.ystep = skip
        comp.parallel_save(result, callback=lambda x: callback_possitions.append(x))

        expected = data1 + data2
        actual = result.read_array(0, 0, 4, 2)

        assert (expected == actual).all()

        assert callback_possitions == expected_steps

@pytest.mark.skipif(yirgacheffe.backends.BACKEND != "NUMPY", reason="Only applies for numpy")
def test_parallel_equality() -> None:
    with tempfile.TemporaryDirectory() as tempdir:
        path1 = os.path.join(tempdir, "test1.tif")
        data1 = np.array([[1, 2, 3, 4], [4, 3, 2, 1]])
        dataset1 = gdal_dataset_with_data((0.0, 0.0), 0.02, data1, filename=path1)
        dataset1.Close()
        with RasterLayer.layer_from_file(path1) as layer1:
            with RasterLayer.empty_raster_layer_like(layer1) as result:
                comp = layer1 == 2
                comp.parallel_save(result)

                expected = np.array([[0, 1, 0, 0], [0, 0, 1, 0]])
                actual = result.read_array(0, 0, 4, 2)

                assert (expected == actual).all()

@pytest.mark.skipif(yirgacheffe.backends.BACKEND != "NUMPY", reason="Only applies for numpy")
def test_parallel_equality_to_file() -> None:
    with tempfile.TemporaryDirectory() as tempdir:
        path1 = os.path.join(tempdir, "test1.tif")
        path2 = os.path.join(tempdir, "result.tif")
        data1 = np.array([[1, 2, 3, 4], [4, 3, 2, 1]])
        dataset1 = gdal_dataset_with_data((0.0, 0.0), 0.02, data1, filename=path1)
        dataset1.Close()
        with RasterLayer.layer_from_file(path1) as layer1:
            comp = layer1 == 2
            with RasterLayer.empty_raster_layer_like(layer1, filename=path2) as result:
                comp.parallel_save(result)
        with RasterLayer.layer_from_file(path2) as actual_result:
            expected = np.array([[0, 1, 0, 0], [0, 0, 1, 0]])
            actual = actual_result.read_array(0, 0, 4, 2)
            assert (expected == actual).all()

@pytest.mark.skipif(yirgacheffe.backends.BACKEND != "NUMPY", reason="Only applies for numpy")
def test_parallel_unary_numpy_apply_with_function() -> None:
    with tempfile.TemporaryDirectory() as tempdir:
        path1 = os.path.join(tempdir, "test1.tif")
        data1 = np.array([[1, 2, 3, 4], [5, 6, 7, 8]])
        dataset1 = gdal_dataset_with_data((0.0, 0.0), 0.02, data1, filename=path1)
        dataset1.Close()
        layer1 = RasterLayer.layer_from_file(path1)

        result = RasterLayer.empty_raster_layer_like(layer1)

        def simple_add(chunk):
            return chunk + 1.0

        comp = layer1.numpy_apply(simple_add)
        comp.ystep = 1
        comp.parallel_save(result)

        expected = data1 + 1.0
        actual = result.read_array(0, 0, 4, 2)

        assert (expected == actual).all()

@pytest.mark.skipif(yirgacheffe.backends.BACKEND != "NUMPY", reason="Only applies for numpy")
def test_parallel_unary_numpy_apply_with_lambda() -> None:
    with tempfile.TemporaryDirectory() as tempdir:
        path1 = os.path.join(tempdir, "test1.tif")
        data1 = np.array([[1, 2, 3, 4], [5, 6, 7, 8]])
        dataset1 = gdal_dataset_with_data((0.0, 0.0), 0.02, data1, filename=path1)
        dataset1.Close()
        layer1 = RasterLayer.layer_from_file(path1)

        result = RasterLayer.empty_raster_layer_like(layer1)

        comp = layer1.numpy_apply(lambda a: a + 1.0)
        comp.ystep = 1
        comp.parallel_save(result)

        expected = data1 + 1.0
        actual = result.read_array(0, 0, 4, 2)

        assert (expected == actual).all()

@pytest.mark.skipif(yirgacheffe.backends.BACKEND != "NUMPY", reason="Only applies for numpy")
def test_parallel_where_simple() -> None:
    with tempfile.TemporaryDirectory() as tempdir:
        path1 = os.path.join(tempdir, "test1.tif")
        data1 = np.array([[0, 1, 0, 2], [0, 0, 1, 1]])
        dataset1 = gdal_dataset_with_data((0.0, 0.0), 0.02, data1, filename=path1)
        dataset1.Close()
        layer1 = RasterLayer.layer_from_file(path1)

        result = RasterLayer.empty_raster_layer_like(layer1)

        comp = LayerOperation.where(layer1 > 0, 1, 2)
        comp.ystep = 1
        comp.parallel_save(result)

        expected = np.where(data1 > 0, 1, 2)
        actual = result.read_array(0, 0, 4, 2)
        assert (expected == actual).all()

@pytest.mark.skipif(yirgacheffe.backends.BACKEND != "NUMPY", reason="Only applies for numpy")
def test_parallel_conv2d() -> None:
    data1 = np.array([
        [0, 0, 0, 0, 0],
        [0, 1, 1, 1, 0],
        [0, 1, 1, 1, 0],
        [0, 1, 1, 1, 0],
        [0, 0, 0, 0, 0],
    ]).astype(np.float32)
    weights = np.array([
        [0.0, 0.1, 0.0],
        [0.1, 0.6, 0.1],
        [0.0, 0.1, 0.0],
    ])

    conv = torch.nn.Conv2d(1, 1, 3, padding=1, bias=False)
    conv.weight = torch.nn.Parameter(torch.from_numpy(np.array([[weights.astype(np.float32)]])))
    tensorres = conv(torch.from_numpy(np.array([[data1]])))
    expected = tensorres.detach().numpy()[0][0]

    with RasterLayer(gdal_dataset_with_data((0.0, 0.0), 0.02, data1)) as layer1:

        calc = layer1.conv2d(weights)
        calc.ystep = 1
        with RasterLayer.empty_raster_layer_like(layer1) as res:
            calc.save(res)
            actual = res.read_array(0, 0, 5, 5)

            # Torch and MLX give slightly different rounding
            assert np.isclose(expected, actual).all()



import os
import tempfile

import numpy as np

from helpers import gdal_dataset_with_data, make_vectors_with_mutlile_ids
from yirgacheffe.layers import ConstantLayer, RasterLayer, VectorLayer, area
from yirgacheffe.window import Area

def test_add_windows() -> None:
	data1 = np.array([[1, 2], [3, 4]])
	data2 = np.array([[10, 20, 30, 40], [50, 60, 70, 80], [90, 100, 110, 120], [130, 140, 150, 160]])

	layer1 = RasterLayer(gdal_dataset_with_data((0.0, 0.0), 0.02, data1))
	layer2 = RasterLayer(gdal_dataset_with_data((0.0, 0.0), 0.02, data2))

	assert layer1.area != layer2.area
	assert layer1.window != layer2.window

	calc = layer1 + layer2

	assert calc.area == layer2.area
	assert calc.window == layer2.window

	result = RasterLayer.empty_raster_layer_like(calc)
	calc.save(result)

	expected = np.array([[11, 22, 30, 40], [53, 64, 70, 80], [90, 100, 110, 120], [130, 140, 150, 160]])
	actual = result.read_array(0, 0, 4, 4)
	assert (expected == actual).all()

def test_multiply_windows() -> None:
	data1 = np.array([[1, 2], [3, 4]])
	data2 = np.array([[10, 20, 30, 40], [50, 60, 70, 80], [90, 100, 110, 120], [130, 140, 150, 160]])

	layer1 = RasterLayer(gdal_dataset_with_data((0.0, 0.0), 0.02, data1))
	layer2 = RasterLayer(gdal_dataset_with_data((0.0, 0.0), 0.02, data2))

	assert layer1.area != layer2.area
	assert layer1.window != layer2.window

	calc = layer1 * layer2

	assert calc.area == layer1.area
	assert calc.window == layer1.window

	result = RasterLayer.empty_raster_layer_like(calc)
	calc.save(result)

	expected = np.array([[10, 40], [150, 240]])
	actual = result.read_array(0, 0, 2, 2)
	assert (expected == actual).all()

def test_add_windows_offset() -> None:
	data1 = np.array([[1, 2], [3, 4]])
	data2 = np.array([[10, 20, 30, 40], [50, 60, 70, 80], [90, 100, 110, 120], [130, 140, 150, 160]])

	layer1 = RasterLayer(gdal_dataset_with_data((-0.02, 0.02), 0.02, data1))
	layer2 = RasterLayer(gdal_dataset_with_data((-0.04, 0.04), 0.02, data2))

	assert layer1.area != layer2.area
	assert layer1.window != layer2.window

	calc = layer1 + layer2

	assert calc.area == layer2.area

	result = RasterLayer.empty_raster_layer_like(calc)
	calc.save(result)

	expected = np.array([[10, 20, 30, 40], [50, 61, 72, 80], [90, 103, 114, 120], [130, 140, 150, 160]])
	actual = result.read_array(0, 0, 4, 4)
	assert (expected == actual).all()

def test_multiply_windows_offset() -> None:
	data1 = np.array([[1, 2], [3, 4]])
	data2 = np.array([[10, 20, 30, 40], [50, 60, 70, 80], [90, 100, 110, 120], [130, 140, 150, 160]])

	layer1 = RasterLayer(gdal_dataset_with_data((-0.02, 0.02), 0.02, data1))
	layer2 = RasterLayer(gdal_dataset_with_data((-0.04, 0.04), 0.02, data2))

	assert layer1.area != layer2.area
	assert layer1.window != layer2.window

	calc = layer1 * layer2

	assert calc.area == layer1.area

	result = RasterLayer.empty_raster_layer_like(calc)
	calc.save(result)

	expected = np.array([[60, 140], [300, 440]])
	actual = result.read_array(0, 0, 2, 2)
	assert (expected == actual).all()

def test_add_windows_sum() -> None:
	data1 = np.array([[1, 2], [3, 4]])
	data2 = np.array([[10, 20, 30, 40], [50, 60, 70, 80], [90, 100, 110, 120], [130, 140, 150, 160]])

	layer1 = RasterLayer(gdal_dataset_with_data((-0.02, 0.02), 0.02, data1))
	layer2 = RasterLayer(gdal_dataset_with_data((-0.04, 0.04), 0.02, data2))

	calc = layer1 + layer2
	total = calc.sum()

	expected = np.array([[10, 20, 30, 40], [50, 61, 72, 80], [90, 103, 114, 120], [130, 140, 150, 160]])
	assert total == np.sum(expected)

def test_multiply_windows_sum() -> None:
	data1 = np.array([[1, 2], [3, 4]])
	data2 = np.array([[10, 20, 30, 40], [50, 60, 70, 80], [90, 100, 110, 120], [130, 140, 150, 160]])

	layer1 = RasterLayer(gdal_dataset_with_data((-0.02, 0.02), 0.02, data1))
	layer2 = RasterLayer(gdal_dataset_with_data((-0.04, 0.04), 0.02, data2))

	calc = layer1 * layer2
	total = calc.sum()

	expected = np.array([[60, 140], [300, 440]])
	assert total == np.sum(expected)

def test_constant_layer_result_rhs_add() -> None:
	data1 = np.array([[1.0, 2.0, 3.0, 4.0], [5.0, 6.0, 7.0, 8.0]])
	layer1 = RasterLayer(gdal_dataset_with_data((0.0, 0.0), 0.02, data1))
	const_layer = ConstantLayer(1.0)

	calc = layer1 + const_layer

	assert calc.area == layer1.area

	result = RasterLayer.empty_raster_layer_like(calc)
	calc.save(result)
	actual = result.read_array(0, 0, 4, 2)

	expected = 1.0 + data1

	assert (expected == actual).all()

def test_constant_layer_result_lhs_add() -> None:
	data1 = np.array([[1.0, 2.0, 3.0, 4.0], [5.0, 6.0, 7.0, 8.0]])
	layer1 = RasterLayer(gdal_dataset_with_data((0.0, 0.0), 0.02, data1))
	const_layer = ConstantLayer(1.0)
	result = RasterLayer.empty_raster_layer_like(layer1)

	intersection = RasterLayer.find_intersection([layer1, const_layer])
	const_layer.set_window_for_intersection(intersection)
	layer1.set_window_for_intersection(intersection)

	calc = const_layer + layer1

	assert calc.area == layer1.area

	result = RasterLayer.empty_raster_layer_like(calc)
	calc.save(result)
	actual = result.read_array(0, 0, 4, 2)

	expected = 1.0 + data1

	assert (expected == actual).all()

def test_constant_layer_result_rhs_multiply() -> None:
	data1 = np.array([[1.0, 2.0, 3.0, 4.0], [5.0, 6.0, 7.0, 8.0]])
	layer1 = RasterLayer(gdal_dataset_with_data((0.0, 0.0), 0.02, data1))
	const_layer = ConstantLayer(2.0)

	calc = layer1 * const_layer

	assert calc.area == layer1.area

	result = RasterLayer.empty_raster_layer_like(calc)
	calc.save(result)
	actual = result.read_array(0, 0, 4, 2)

	expected = data1 * 2.0

	assert (expected == actual).all()

def test_constant_layer_result_lhs_multiply() -> None:
	data1 = np.array([[1.0, 2.0, 3.0, 4.0], [5.0, 6.0, 7.0, 8.0]])
	layer1 = RasterLayer(gdal_dataset_with_data((0.0, 0.0), 0.02, data1))
	const_layer = ConstantLayer(2.0)
	result = RasterLayer.empty_raster_layer_like(layer1)

	intersection = RasterLayer.find_intersection([layer1, const_layer])
	const_layer.set_window_for_intersection(intersection)
	layer1.set_window_for_intersection(intersection)

	calc = const_layer * layer1

	assert calc.area == layer1.area

	result = RasterLayer.empty_raster_layer_like(calc)
	calc.save(result)
	actual = result.read_array(0, 0, 4, 2)

	expected = 2.0 * data1

	assert (expected == actual).all()

def test_vector_layers_add() -> None:
	data1 = np.array([[1, 2], [3, 4]])
	layer1 = RasterLayer(gdal_dataset_with_data((0.0, 0.0), 1.0, data1))

	with tempfile.TemporaryDirectory() as tempdir:
		path = os.path.join(tempdir, "test.gpkg")
		areas = {
			(Area(-10.0, 10.0, 0.0, 0.0), 42),
			(Area(0.0, 0.0, 10, -10), 43)
		}
		make_vectors_with_mutlile_ids(areas, path)

		burn_value = 2
		layer2 = VectorLayer.layer_from_file(path, None, layer1.pixel_scale, layer1.projection, burn_value=burn_value)
		layer2_total = layer2.sum()
		assert layer2_total == ((layer2.window.xsize * layer2.window.ysize) / 2) * burn_value

		calc = layer1 + layer2

		assert calc.area == layer2.area

		total = calc.sum()
		assert total == layer2_total + np.sum(data1)

def test_vector_layers_multiply() -> None:
	data1 = np.array([[1, 2], [3, 4]])
	layer1 = RasterLayer(gdal_dataset_with_data((-1.0, 1.0), 1.0, data1))

	with tempfile.TemporaryDirectory() as tempdir:
		path = os.path.join(tempdir, "test.gpkg")
		areas = {
			(Area(-10.0, 10.0, 0.0, 0.0), 42),
			(Area(0.0, 0.0, 10, -10), 43)
		}
		make_vectors_with_mutlile_ids(areas, path)

		burn_value = 2
		layer2 = VectorLayer.layer_from_file(path, None, layer1.pixel_scale, layer1.projection, burn_value=burn_value)
		layer2_total = layer2.sum()
		assert layer2_total == ((layer2.window.xsize * layer2.window.ysize) / 2) * burn_value

		calc = layer1 * layer2

		assert calc.area == layer1.area

		result = RasterLayer.empty_raster_layer_like(calc)
		calc.save(result)
		actual = result.read_array(0, 0, 2, 2)

		expected = np.array([[2, 0], [0, 8]])
		assert (expected == actual).all()



import numpy as np

from helpers import gdal_dataset_of_region, gdal_dataset_with_data
from yirgacheffe.layers import RasterLayer, RescaledRasterLayer
from yirgacheffe.window import Area, PixelScale, Window


def test_simple_scale_down() -> None:
    area = Area(-10, 10, 10, -10)
    dataset = gdal_dataset_of_region(area, 0.02)
    with RasterLayer(dataset) as raster:
        target_pixel_scale = PixelScale(0.01, -0.01)
        with RescaledRasterLayer(raster, target_pixel_scale) as layer:
            assert layer.area == area
            assert layer.pixel_scale == target_pixel_scale
            assert layer.geo_transform == (-10, 0.01, 0.0, 10, 0.0, -0.01)
            assert layer.window == Window(0, 0, 2000, 2000)

def test_simple_scale_up() -> None:
    area = Area(-10, 10, 10, -10)
    dataset = gdal_dataset_of_region(area, 0.02)
    with RasterLayer(dataset) as raster:
        target_pixel_scale = PixelScale(0.04, -0.04)
        with RescaledRasterLayer(raster, target_pixel_scale) as layer:
            assert layer.area == area
            assert layer.pixel_scale == target_pixel_scale
            assert layer.geo_transform == (-10, 0.04, 0.0, 10, 0.0, -0.04)
            assert layer.window == Window(0, 0, 500, 500)

def test_scaling_up_pixels() -> None:
    # data has top left and bottom right quarters as 0
    # and the remaining quarters as 1
    data = np.zeros((4, 4))
    data[0:2,2:4] = 1
    data[2:4,0:2] = 1
    dataset = gdal_dataset_with_data((0, 0), 1.0, data)
    with RasterLayer(dataset) as raster:

        target_pixel_scale = PixelScale(0.5, -0.5)
        with RescaledRasterLayer(raster, target_pixel_scale) as layer:
            assert layer.area == Area(0, 0, 4, -4)
            assert layer.pixel_scale == target_pixel_scale
            assert layer.geo_transform == (0.0, 0.5, 0.0, 0.0, 0.0, -0.5)
            assert layer.window == Window(0, 0, 8, 8)

            actual_raster = layer.read_array(0, 0, 8, 8)
            expected_raster = np.zeros((8, 8))
            expected_raster[0:4,4:8] = 1
            expected_raster[4:8,0:4] = 1
            assert (expected_raster == actual_raster).all()

            # Try getting just the quads
            expected_quad_raster = np.zeros((4, 4))
            actual_raster = layer.read_array(0, 0, 4, 4)
            assert (expected_quad_raster == actual_raster).all()
            actual_raster = layer.read_array(4, 4, 4, 4)
            assert (expected_quad_raster == actual_raster).all()

            expected_quad_raster = np.ones((4, 4))
            actual_raster = layer.read_array(0, 4, 4, 4)
            assert (expected_quad_raster == actual_raster).all()
            actual_raster = layer.read_array(4, 0, 4, 4)
            assert (expected_quad_raster == actual_raster).all()

            for box in range(1, 7):
                # anchored top left
                actual_raster = layer.read_array(0, 0, box, box)
                exepected_sub_raster = expected_raster[0:box, 0:box]
                assert (exepected_sub_raster == actual_raster).all()

                # anchored bottom right
                actual_raster = layer.read_array(box, box, 8 - box, 8 - box)
                exepected_sub_raster = expected_raster[box:8, box:8]
                assert (exepected_sub_raster == actual_raster).all()

                # anchored top right
                actual_raster = layer.read_array(box, 0, 8 - box, box)
                exepected_sub_raster = expected_raster[0:box, box:8]
                assert (exepected_sub_raster == actual_raster).all()

                # anchored bottom left
                actual_raster = layer.read_array(0, box,  box, 8 - box)
                exepected_sub_raster = expected_raster[box:8, 0:box]
                assert (exepected_sub_raster == actual_raster).all()

                # columns
                actual_raster = layer.read_array(box, 0, 1, 8)
                exepected_sub_raster = expected_raster[0:8, box:box+1]
                assert (exepected_sub_raster == actual_raster).all()

                # rows
                actual_raster = layer.read_array(0, box, 8, 1)
                exepected_sub_raster = expected_raster[box:box+1, 0:8]
                assert (exepected_sub_raster == actual_raster).all()

def test_scaling_down_pixels() -> None:
    # data has top left and bottom right quarters as 0
    # and the remaining quarters as 1
    data = np.zeros((8, 8))
    data[0:4,4:8] = 1
    data[4:8,0:4] = 1
    dataset = gdal_dataset_with_data((0, 0), 1.0, data)
    with RasterLayer(dataset) as raster:

        target_pixel_scale = PixelScale(2.0, -2.0)
        with RescaledRasterLayer(raster, target_pixel_scale) as layer:
            assert layer.area == Area(0, 0, 8, -8)
            assert layer.pixel_scale == target_pixel_scale
            assert layer.geo_transform == (0.0, 2.0, 0.0, 0.0, 0.0, -2.0)
            assert layer.window == Window(0, 0, 4, 4)

            actual_raster = layer.read_array(0, 0, 4, 4)
            expected_raster = np.zeros((4, 4))
            expected_raster[0:2,2:4] = 1
            expected_raster[2:4,0:2] = 1
            assert (expected_raster == actual_raster).all()

            # Try getting just the quads
            expected_quad_raster = np.zeros((2, 2))
            actual_raster = layer.read_array(0, 0, 2, 2)
            assert (expected_quad_raster == actual_raster).all()
            actual_raster = layer.read_array(2, 2, 2, 2)
            assert (expected_quad_raster == actual_raster).all()

            expected_quad_raster = np.ones((2, 2))
            actual_raster = layer.read_array(0, 2, 2, 2)
            assert (expected_quad_raster == actual_raster).all()
            actual_raster = layer.read_array(2, 0, 2, 2)
            assert (expected_quad_raster == actual_raster).all()

            for box in range(1, 3):
                # anchored top left
                actual_raster = layer.read_array(0, 0, box, box)
                exepected_sub_raster = expected_raster[0:box, 0:box]
                assert (exepected_sub_raster == actual_raster).all()

                # anchored bottom right
                actual_raster = layer.read_array(box, box, 4 - box, 4 - box)
                exepected_sub_raster = expected_raster[box:4, box:4]
                assert (exepected_sub_raster == actual_raster).all()

                # anchored top right
                actual_raster = layer.read_array(box, 0, 4 - box, box)
                exepected_sub_raster = expected_raster[0:box, box:4]
                assert (exepected_sub_raster == actual_raster).all()

                # anchored bottom left
                actual_raster = layer.read_array(0, box,  box, 4 - box)
                exepected_sub_raster = expected_raster[box:4, 0:box]
                assert (exepected_sub_raster == actual_raster).all()

                # columns
                actual_raster = layer.read_array(box, 0, 1, 4)
                exepected_sub_raster = expected_raster[0:4, box:box+1]
                assert (exepected_sub_raster == actual_raster).all()

                # rows
                actual_raster = layer.read_array(0, box, 4, 1)
                exepected_sub_raster = expected_raster[box:box+1, 0:4]
                assert (exepected_sub_raster == actual_raster).all()

def test_rescaled_up_in_operation() -> None:
    data1 = np.zeros((8, 8))
    data1[0:4,4:8] = 1
    data1[4:8,0:4] = 1
    dataset1 = gdal_dataset_with_data((0, 0), 1.0, data1)
    raster1 = RasterLayer(dataset1)

    data2 = np.zeros((4, 4))
    data2[0:2,0:2] = 1
    data2[2:4,2:4] = 1
    dataset2 = gdal_dataset_with_data((0, 0), 2.0, data2)
    raster2 = RasterLayer(dataset2)

    rescaled = RescaledRasterLayer(raster2, raster1.pixel_scale)

    assert raster1.window == rescaled.window
    assert raster1.area == rescaled.area

    calc = raster1 + rescaled
    calc.ystep = 1
    assert calc.sum() == (8 * 8)

def test_rescaled_down_in_operation() -> None:
    data1 = np.zeros((8, 8))
    data1[0:4,4:8] = 1
    data1[4:8,0:4] = 1
    dataset1 = gdal_dataset_with_data((0, 0), 1.0, data1)
    raster1 = RasterLayer(dataset1)

    data2 = np.zeros((4, 4))
    data2[0:2,0:2] = 1
    data2[2:4,2:4] = 1
    dataset2 = gdal_dataset_with_data((0, 0), 2.0, data2)
    raster2 = RasterLayer(dataset2)

    rescaled = RescaledRasterLayer(raster1, raster2.pixel_scale)

    assert raster2.window == rescaled.window
    assert raster2.area == rescaled.area

    calc = rescaled + raster2
    calc.ystep = 1
    assert calc.sum() == (4 * 4)

def test_rescaled_up_with_window_set() -> None:
    # data has top left and bottom right quarters as 0
    # and the remaining quarters as 1
    data = np.zeros((4, 4))
    data[0:2,2:4] = 1
    data[2:4,0:2] = 1
    dataset = gdal_dataset_with_data((0, 0), 1.0, data)
    with RasterLayer(dataset) as raster:

        target_pixel_scale = PixelScale(0.5, -0.5)
        with RescaledRasterLayer(raster, target_pixel_scale) as layer:
            assert layer.area == Area(0.0, 0.0, 4.0, -4.0)

            layer.set_window_for_intersection(Area(1.0, -1.0, 3.0, -3.0))
            assert layer.area == Area(1.0, -1.0, 3.0, -3.0)
            assert layer.window == Window(2, 2, 4, 4)

            actual_raster = layer.read_array(0, 0, 4, 4)
            expected_raster = np.zeros((4, 4))
            expected_raster[0:2,2:4] = 1
            expected_raster[2:4,0:2] = 1

            assert (actual_raster == expected_raster).all()

def test_rescaled_up_with_window_set_2() -> None:
    # data has top left and bottom right quarters as 0
    # and the remaining quarters as 1
    data = np.zeros((4, 4))
    data[0:2,2:4] = 1
    data[2:4,0:2] = 1
    dataset = gdal_dataset_with_data((0, 0), 1.0, data)
    with RasterLayer(dataset) as raster:

        target_pixel_scale = PixelScale(0.5, -0.5)
        with RescaledRasterLayer(raster, target_pixel_scale) as layer:
            assert layer.area == Area(0.0, 0.0, 4.0, -4.0)

            expected_raster = np.zeros((6, 6))
            expected_raster[0:3,3:6] = 1
            expected_raster[3:6,0:3] = 1

            # Try get the intended data without the window offset first
            actual_raster = layer.read_array(1, 1, 6, 6)
            assert (actual_raster == expected_raster).all()

            layer.set_window_for_intersection(Area(0.5, -0.5, 3.5, -3.5))
            assert layer.area == Area(0.5, -0.5, 3.5, -3.5)
            assert layer.window == Window(1, 1, 6, 6)

            actual_raster = layer.read_array(0, 0, 6, 6)
            assert (actual_raster == expected_raster).all()

def test_rescaled_down_with_window_set() -> None:
    # data has top left and bottom right quarters as 0
    # and the remaining quarters as 1
    data = np.zeros((8, 8))
    data[0:4,4:8] = 1
    data[4:8,0:4] = 1
    dataset = gdal_dataset_with_data((0, 0), 1.0, data)
    with RasterLayer(dataset) as raster:

        target_pixel_scale = PixelScale(2.0, -2.0)
        with RescaledRasterLayer(raster, target_pixel_scale) as layer:
            assert layer.area == Area(0.0, 0.0, 8.0, -8.0)

            layer.set_window_for_intersection(Area(2.0, -2.0, 6.0, -6.0))
            assert layer.area == Area(2.0, -2.0, 6.0, -6.0)
            assert layer.window == Window(1, 1, 2, 2)

            actual_raster = layer.read_array(0, 0, 2, 2)
            expected_raster = np.zeros((2, 2))
            expected_raster[0:1,1:2] = 1
            expected_raster[1:2,0:1] = 1

            assert (actual_raster == expected_raster).all()
